@@ -1,6 +1,7 @@
+from __future__ import absolute_import
 from sfepy.linalg import norm_l2_along_axis
 
-from quantum_common import common
+from examples.quantum.quantum_common import common
 
 def fun_v(ts, coor, mode=None, **kwargs):
     if not mode == 'qp': return
