@@ -32,13 +32,15 @@ Additional steps for sfepy.org deployment:
 - copy doc/_build/html/ to <sfepy.org>/doc-devel/
 - copy gallery/gallery.html and gallery/images/ to <sfepy.org>/
 """
+from __future__ import absolute_import
 import sys
-sys.path.append( '.' )
+import six
+sys.path.append('.')
 import os
 import tempfile
 import glob
 import re
-from optparse import OptionParser
+from argparse import ArgumentParser, RawDescriptionHelpFormatter
 
 import matplotlib.image as image
 
@@ -51,7 +53,6 @@ from sfepy.postprocess.domain_specific import DomainSpecificPlot
 
 omits = [
     'vibro_acoustic3d_mid.py',
-    'linear_elastic_mM.py',
     'its2D_5.py',
     'linear_elastic_probes.py',
     '__init__.py',
@@ -196,6 +197,18 @@ custom = {
             'opacity' : {'wireframe' : 0.2},
         },
     },
+    'linear_elasticity/shell10x_cantilever.py' : {
+        '' : {
+            'is_wireframe' : True,
+            'domain_specific' : {
+                'u_disp' : DomainSpecificPlot('plot_displacements',
+                                              ['rel_scaling=1']),
+            },
+            'view' : (-45, 81, 0.59, [-0.075,  0.023,  0.093]),
+            'roll' : -75.0,
+            'opacity' : {'wireframe' : 0.5},
+        },
+    },
     'navier_stokes/stokes_slip_bc.py' : {
         '' : {
             'view' : (-63, 52, 5.2, [-1.5, -0.65, 0.12]),
@@ -205,6 +218,15 @@ custom = {
             'rel_scaling' : 0.1,
         },
     },
+    'navier_stokes/stokes_slip_bc_penalty.py' : {
+        '' : {
+            'view' : (-63, 52, 5.2, [-1.5, -0.65, 0.12]),
+            'roll' : -32,
+            'resolution' : (800, 600),
+            'layout' : 'col',
+            'rel_scaling' : 0.1,
+        },
+    },
     'multi_physics/thermo_elasticity_ess.py' : {
         '' : {
             'is_wireframe' : True,
@@ -279,6 +301,7 @@ def generate_images(images_dir, examples_dir):
     from sfepy.applications import solve_pde
     from sfepy.postprocess.viewer import Viewer
     from sfepy.postprocess.utils import mlab
+    from sfepy.solvers.ts_solvers import StationarySolver
 
     prefix = output.prefix
 
@@ -323,8 +346,8 @@ def generate_images(images_dir, examples_dir):
             else:
                 views = default_views
 
-            tsolver = problem.get_time_solver()
-            if tsolver.ts is None:
+            tsolver = problem.get_solver()
+            if isinstance(tsolver, StationarySolver):
                 suffix = None
 
             else:
@@ -332,7 +355,7 @@ def generate_images(images_dir, examples_dir):
 
             filename = problem.get_output_name(suffix=suffix)
 
-            for suffix, kwargs in views.iteritems():
+            for suffix, kwargs in six.iteritems(views):
                 fig_filename = _get_fig_filename(ebase, images_dir, suffix)
 
                 fname = edit_filename(filename, suffix=suffix)
@@ -427,8 +450,8 @@ def generate_rst_files(rst_dir, examples_dir, images_dir):
 
         dir_map.setdefault(base_dir, []).append((ex_filename, rst_filename))
 
-    for dirname, filenames in dir_map.iteritems():
-        filenames = sorted(filenames, cmp=lambda a, b: cmp(a[1], b[1]))
+    for dirname, filenames in six.iteritems(dir_map):
+        filenames = sorted(filenames, key=lambda a: a[1])
         dir_map[dirname ] = filenames
 
     # Main index.
@@ -455,7 +478,7 @@ def generate_rst_files(rst_dir, examples_dir, images_dir):
 
             ifd.write('    %s\n' % rst_filename_ns)
             fig_include = ''
-            fig_base = _get_fig_filenames(ebase, images_dir).next()
+            fig_base = next(_get_fig_filenames(ebase, images_dir))
             for fig_filename in _get_fig_filenames(ebase, images_dir):
                 rst_fig_filename = _make_sphinx_path(fig_filename)
 
@@ -546,7 +569,7 @@ def generate_gallery_html(examples_dir, output_filename, gallery_dir,
             link = os.path.join(link_prefix,
                                 os.path.splitext(link_base)[0] + '.html')
 
-            _get_fig_filenames(ebase, thumbnails_dir).next()
+            next(_get_fig_filenames(ebase, thumbnails_dir))
             for thumbnail_filename in _get_fig_filenames(ebase,
                                                          thumbnails_dir):
                 if not os.path.isfile(thumbnail_filename):
@@ -576,42 +599,42 @@ def generate_gallery_html(examples_dir, output_filename, gallery_dir,
 
     output('...done')
 
-usage = '%prog [options]\n' + __doc__.rstrip()
-
-help = {
+helps = {
     'examples_dir' :
-    'directory containing examples [default: %default]',
+    'directory containing examples [default: %(default)s]',
     'images_dir' :
     'directory where to store gallery images [default: gallery/images]',
     'no_images' :
     'do not (re)generate images and thumbnails',
     'output_filename' :
-    'output file name [default: %default]',
+    'output file name [default: %(default)s]',
     'link_prefix' :
     'prefix to be prepended to links to examples pages in gallery '
-    '[default: %default]',
+    '[default: %(default)s]',
 }
 
 def main():
-    parser = OptionParser(usage=usage, version='%prog')
-    parser.add_option('-e', '--examples-dir', metavar='directory',
-                      action='store', dest='examples_dir',
-                      default='examples', help=help['examples_dir'])
-    parser.add_option('-i', '--images-dir', metavar='directory',
-                      action='store', dest='images_dir',
-                      default=None, help=help['images_dir'])
-    parser.add_option('-n', '--no-images',
-                      action='store_true', dest='no_images',
-                      default=False, help=help['no_images'])
-    parser.add_option('-o', '--output', metavar='output_filename',
-                      action='store', dest='output_filename',
-                      default='gallery/gallery.html',
-                      help=help['output_filename'])
-    parser.add_option('-l', '--link-prefix', metavar='prefix',
-                      action='store', dest='link_prefix',
-                      default='http://sfepy.org/doc-devel',
-                      help=help['link_prefix'])
-    (options, args) = parser.parse_args()
+    parser = ArgumentParser(description=__doc__,
+                            formatter_class=RawDescriptionHelpFormatter)
+    parser.add_argument('--version', action='version', version='%(prog)s')
+    parser.add_argument('-e', '--examples-dir', metavar='directory',
+                        action='store', dest='examples_dir',
+                        default='examples', help=helps['examples_dir'])
+    parser.add_argument('-i', '--images-dir', metavar='directory',
+                        action='store', dest='images_dir',
+                        default=None, help=helps['images_dir'])
+    parser.add_argument('-n', '--no-images',
+                        action='store_true', dest='no_images',
+                        default=False, help=helps['no_images'])
+    parser.add_argument('-o', '--output', metavar='output_filename',
+                        action='store', dest='output_filename',
+                        default='gallery/gallery.html',
+                        help=helps['output_filename'])
+    parser.add_argument('-l', '--link-prefix', metavar='prefix',
+                        action='store', dest='link_prefix',
+                        default='http://sfepy.org/doc-devel',
+                        help=helps['link_prefix'])
+    options = parser.parse_args()
 
     examples_dir = os.path.realpath(options.examples_dir)
 
