@@ -1,8 +1,11 @@
+from __future__ import absolute_import
 import time
 
 import numpy as nm
 
 from sfepy.base.base import output, iter_dict_of_lists, Struct, basestr
+import six
+
 
 def parse_approx_order(approx_order):
     """
@@ -51,8 +54,8 @@ def parse_shape(shape, dim):
         except KeyError:
             raise ValueError('unsupported field shape! (%s)', shape)
 
-    elif isinstance(shape, int):
-        shape = (shape,)
+    elif isinstance(shape, six.integer_types):
+        shape = (int(shape),)
 
     return shape
 
@@ -67,7 +70,7 @@ def setup_extra_data(conn_info):
 
 def fields_from_conf(conf, regions):
     fields = {}
-    for key, val in conf.iteritems():
+    for key, val in six.iteritems(conf):
         field = Field.from_conf(val, regions)
         fields[field.name] = field
 
@@ -121,13 +124,14 @@ class Field(Struct):
         Create a Field subclass instance based on the configuration.
         """
         if Field._all is None:
-            import sfepy
+            from sfepy import get_paths
             from sfepy.base.base import load_classes
 
             field_files = [ii for ii
-                           in sfepy.get_paths('sfepy/discrete/fem/fields*.py')
-                           if 'fields_base.py' not in ii] \
-                           + sfepy.get_paths('sfepy/discrete/iga/fields*.py')
+                           in get_paths('sfepy/discrete/fem/fields*.py')
+                           if 'fields_base.py' not in ii]
+            field_files += get_paths('sfepy/discrete/iga/fields*.py')
+            field_files += get_paths('sfepy/discrete/structural/fields*.py')
             Field._all = load_classes(field_files, [Field], ignore_errors=True,
                                       name_attr='family_name')
         table = Field._all
@@ -177,13 +181,24 @@ class Field(Struct):
         """
         self.mappings = {}
         if clear_all:
-            self.mappings0 = {}
+            if hasattr(self, 'mappings0'):
+                self.mappings0.clear()
+            else:
+                self.mappings0 = {}
 
     def save_mappings(self):
         """
         Save current reference mappings to `mappings0` attribute.
         """
-        self.mappings0 = self.mappings.copy()
+        import sfepy.base.multiproc as multi
+
+        if multi.is_remote_dict(self.mappings0):
+            for k, v in six.iteritems(self.mappings):
+                m, _ = self.mappings[k]
+                nv = (m.bf, m.bfg, m.det, m.volume, m.normal)
+                self.mappings0[k] = nv
+        else:
+            self.mappings0 = self.mappings.copy()
 
     def get_mapping(self, region, integral, integration,
                     get_saved=False, return_key=False):
@@ -209,11 +224,18 @@ class Field(Struct):
         key : tuple
             The key of the mapping in `mappings` or `mappings0`.
         """
+        import sfepy.base.multiproc as multi
+
         key = (region.name, integral.order, integration)
 
         if get_saved:
             out = self.mappings0.get(key, None)
-
+            if multi.is_remote_dict(self.mappings0) and out is not None:
+                m, i = self.create_mapping(region, integral, integration)
+                m.bf[:], m.bfg[:], m.det[:], m.volume[:] = out[0:4]
+                if m.normal is not None:
+                    m.normal[:] = m[4]
+                out = m, i
         else:
             out = self.mappings.get(key, None)
 
@@ -304,6 +326,7 @@ class Field(Struct):
         """
         from sfepy.discrete.common.global_interp import get_ref_coors
         from sfepy.discrete.common.extmods.crefcoors import evaluate_in_rc
+        from sfepy.base.base import complex_types
 
         output('evaluating in %d points...' % coors.shape[0], verbose=verbose)
 
@@ -317,21 +340,34 @@ class Field(Struct):
         tt = time.clock()
 
         # Interpolate to the reference coordinates.
+        source_dtype = nm.float64 if source_vals.dtype in complex_types\
+            else source_vals.dtype
         if mode == 'val':
             vals = nm.empty((coors.shape[0], source_vals.shape[1], 1),
-                            dtype=source_vals.dtype)
+                            dtype=source_dtype)
             cmode = 0
 
         elif mode == 'grad':
             vals = nm.empty((coors.shape[0], source_vals.shape[1],
                              coors.shape[1]),
-                            dtype=source_vals.dtype)
+                            dtype=source_dtype)
             cmode = 1
 
         ctx = self.create_basis_context()
 
-        evaluate_in_rc(vals, ref_coors, cells, status, source_vals,
-                       self.get_econn('volume', self.region), cmode, ctx)
+        if source_vals.dtype in complex_types:
+            valsi = vals.copy()
+            evaluate_in_rc(vals, ref_coors, cells, status,
+                           nm.ascontiguousarray(source_vals.real),
+                           self.get_econn('volume', self.region), cmode, ctx)
+            evaluate_in_rc(valsi, ref_coors, cells, status,
+                           nm.ascontiguousarray(source_vals.imag),
+                           self.get_econn('volume', self.region), cmode, ctx)
+            vals = vals + valsi * 1j
+        else:
+            evaluate_in_rc(vals, ref_coors, cells, status, source_vals,
+                           self.get_econn('volume', self.region), cmode, ctx)
+
         output('interpolation: %f s' % (time.clock()-tt),verbose=verbose)
 
         output('...done',verbose=verbose)
