@@ -1,11 +1,14 @@
+from __future__ import absolute_import
 import re
 from copy import copy
 
 import numpy as nm
 
-from sfepy.base.base import output, assert_, Struct
+from sfepy.base.base import assert_, Struct
+import six
+from six.moves import range
 
-_depends = re.compile('r\.([a-zA-Z_0-9.]+)').findall
+_depends = re.compile('r\.([a-zA-Z_\-0-9.]+)').findall
 
 def get_parents(selector):
     """
@@ -22,14 +25,14 @@ def get_dependency_graph(region_defs):
     """
     graph = {}
     name_to_sort_name = {}
-    for sort_name, rdef in region_defs.iteritems():
+    for sort_name, rdef in six.iteritems(region_defs):
         name, sel = rdef.name, rdef.select
-        if name_to_sort_name.has_key(name):
+        if name in name_to_sort_name:
             msg = 'region %s/%s already defined!' % (sort_name, name)
             raise ValueError(msg)
         name_to_sort_name[name] = sort_name
 
-        if not graph.has_key(name):
+        if name not in graph:
             graph[name] = [0]
 
         for parent in get_parents(sel):
@@ -49,7 +52,7 @@ def sort_by_dependency(graph):
     while idone < n_nod:
 
         dep_removed = 0
-        for node, deps in graph.iteritems():
+        for node, deps in six.iteritems(graph):
 
             if (len(deps) == 1) and not deps[0]:
                 out.append(node)
@@ -69,7 +72,7 @@ def sort_by_dependency(graph):
                         dep_removed += 1
 
         if (idone <= idone0) and not dep_removed:
-            raise ValueError, 'circular dependency'
+            raise ValueError('circular dependency')
         idone0 = idone
 
     return out
@@ -86,12 +89,6 @@ def are_disjoint(r1, r2):
 def _join(def1, op, def2):
     return '(' + def1 + ' ' + op + ' ' + def2 + ')'
 
-def _try_delete(obj, ig):
-    try:
-        del obj[ig]
-    except KeyError:
-        pass
-
 class Region(Struct):
     """
     Region defines a subset of a FE domain.
@@ -110,11 +107,6 @@ class Region(Struct):
     followed by one of ('v', 'e', 'f', 'c', and 's') for vertices, edges,
     faces, cells, and facets.
 
-    Notes
-    -----
-    Functions depending on `ig` are adapters for current code that should be
-    removed after new assembling is done.
-
     Created: 31.10.2005
     """
     __can = {
@@ -252,8 +244,7 @@ class Region(Struct):
                         tdim=tdim, kind_tdim=None,
                         entities=[None] * (tdim + 1),
                         kind=None, parent=parent, shape=None,
-                        mirror_region=None, ig_map=None,
-                        ig_map_i=None)
+                        mirror_region=None, is_empty=False)
         self.set_kind(kind)
 
     def set_kind(self, kind):
@@ -289,8 +280,6 @@ class Region(Struct):
             if not ican:
                 self.entities[ii] = nm.empty(0, dtype=nm.uint32)
 
-        self._igs = None
-
         self.set_kind_tdim()
 
     def set_kind_tdim(self):
@@ -428,7 +417,7 @@ class Region(Struct):
         else:
             self.cells
 
-    def setup_from_highest(self, dim, allow_lower=True):
+    def setup_from_highest(self, dim, allow_lower=True, allow_empty=False):
         """
         Setup entities of topological dimension `dim` using the available
         entities of the highest topological dimension.
@@ -441,18 +430,33 @@ class Region(Struct):
                     break
 
         else:
-            msg = 'region "%s" has no entities!'
-            raise ValueError(msg % self.name)
+            if not (allow_empty or self.is_empty):
+                msg = 'region "%s" has no entities!'
+                raise ValueError(msg % self.name)
+
+            if self.entities[dim] is None:
+                self.entities[dim] = nm.empty(0, dtype=nm.uint32)
+
+            self.is_empty = True
+
+            return
 
         cmesh = self.domain.cmesh
         if idim <= dim:
-            if not allow_lower:
+            if not (allow_lower or allow_empty):
                 msg = 'setup_from_highest() can be used only with dim < %d'
                 raise ValueError(msg % idim)
 
-            cmesh.setup_connectivity(dim, idim)
-            ents = self.get_entities(idim)
-            self.entities[dim] = cmesh.get_complete(dim, ents, idim)
+            if allow_lower:
+                cmesh.setup_connectivity(dim, idim)
+                ents = self.get_entities(idim)
+                self.entities[dim] = cmesh.get_complete(dim, ents, idim)
+
+            else:
+                for idim in range(self.kind_tdim - 1, -1, -1):
+                    self.entities[idim] = nm.empty(0, dtype=nm.uint32)
+
+                self.is_empty = True
 
         else:
             cmesh.setup_connectivity(idim, dim)
@@ -472,20 +476,24 @@ class Region(Struct):
         vv = self.vertices
         self.entities[dim] = cmesh.get_complete(dim, vv, 0)
 
-    def finalize(self):
+    def finalize(self, allow_empty=False):
         """
         Initialize the entities corresponding to the region kind and regenerate
         all already existing (accessed) entities of lower topological dimension
         from the kind entities.
         """
         self._access(self.kind_tdim)
+        if allow_empty and ('cell' in self.true_kind):
+            self.is_empty = self.cells.shape[0] == 0
+
         for idim in range(self.kind_tdim - 1, -1, -1):
             if self.can[idim] and self.entities[idim] is not None:
                 try:
-                    self.setup_from_highest(idim, allow_lower=False)
+                    self.setup_from_highest(idim, allow_lower=False,
+                                            allow_empty=allow_empty)
 
-                except ValueError, exc:
-                    msg = '\n'.join((exc.message,
+                except ValueError as exc:
+                    msg = '\n'.join((str(exc),
                                      'fix region kind? (region: %s, kind: %s)'
                                      % (self.name, self.kind)))
                     raise ValueError(msg)
@@ -541,128 +549,45 @@ class Region(Struct):
     def delete_zero_faces(self, eps=1e-14):
         raise NotImplementedError
 
-    @property
-    def igs(self):
-        """
-        Cell group indices according to region kind.
-        """
-        if self.parent is not None:
-            self._igs = self.domain.regions[self.parent].igs
-
-        elif self._igs is None:
-            if 'vertex' in self.true_kind:
-                self._igs = self.domain.cmesh.get_igs(self.vertices, 0)
-
-            elif 'edge' in self.true_kind:
-                self._igs = self.domain.cmesh.get_igs(self.edges, 1)
-
-            elif 'face' in self.true_kind:
-                self._igs = self.domain.cmesh.get_igs(self.faces, 2)
-
-            elif 'cell' in self.true_kind:
-                self._igs = self.domain.cmesh.get_igs(self.cells, self.tdim)
-
-            if not len(self._igs):
-                output('warning: region %s of %s kind has empty group indices!'
-                       % (self.name, self.kind))
-
-        return self._igs
-
     def update_shape(self):
         """
         Update shape of each group according to region vertices, edges,
         faces and cells.
         """
-        get = self.domain.cmesh.get_from_cell_group
+        n_vertex = self.vertices.shape[0]
+        n_cell = self.cells.shape[0]
+        n_edge = self.edges.shape[0] if self.tdim > 1 else 0
+        n_face = self.faces.shape[0] if self.tdim == 3 else 0
+        n_facet = self.facets.shape[0] if self.tdim > 1 else 0
 
-        self.shape = {}
-        for ig in self.igs:
-            n_vertex = get(ig, 0, self.vertices).shape[0]
-            n_cell = get(ig, self.tdim, self.cells).shape[0]
+        self.shape = Struct(n_vertex=n_vertex,
+                            n_edge=n_edge,
+                            n_face=n_face,
+                            n_facet=n_facet,
+                            n_cell=n_cell)
 
-            if self.tdim > 1:
-                n_edge = get(ig, 1, self.edges).shape[0]
-
-            else:
-                n_edge = 0
-
-            if self.tdim == 3:
-                n_face = get(ig, 2, self.faces).shape[0]
-
-            else:
-                n_face = 0
-
-            self.shape[ig] = Struct(n_vertex=n_vertex,
-                                    n_edge=n_edge,
-                                    n_face=n_face,
-                                    n_cell=n_cell)
-
-    def get_entities(self, dim, ig=None):
+    def get_entities(self, dim):
         """
-        Return mesh entities of dimension `dim`, and optionally with the cell
-        group `ig`.
+        Return mesh entities of dimension `dim`.
         """
-        if ig is not None:
-            out = self.domain.cmesh.get_from_cell_group(ig, dim,
-                                                        self.entities[dim])
+        if dim <= self.tdim:
+            self._access(dim)
+            out = self.entities[dim]
 
         else:
-            if dim <= self.tdim:
-                self._access(dim)
-                out = self.entities[dim]
-
-            else:
-                out = nm.empty(0, dtype=nm.uint32)
+            out = nm.empty(0, dtype=nm.uint32)
 
         return out
 
-    def get_vertices_of_cells(self):
-        """
-        Return all vertices, that are in some cell of the region.
-        """
-        vertices = self.domain.cmesh.get_incident(0, self.cells, self.tdim)
-
-        return nm.unique(vertices)
-
-    def get_vertices(self, ig):
-        out = self.domain.cmesh.get_from_cell_group(ig, 0, self.vertices)
-        return out
-
-    def get_edges(self, ig):
-        out = self.domain.cmesh.get_from_cell_group(ig, 1, self.edges)
-        return out
-
-    def get_faces(self, ig):
-        out = self.domain.cmesh.get_from_cell_group(ig, 2, self.faces)
-        return out
-
-    def get_facets(self, ig):
-        """
-        Return either region vertices (in 1D), edges (in 2D) or faces (in 3D).
-        """
-        if self.tdim == 1:
-            return self.get_vertices(ig)
-
-        elif self.tdim == 2:
-            return self.get_edges(ig)
-
-        else:
-            return self.get_faces(ig)
-
-    def get_cells(self, ig, true_cells_only=True, offset=True):
+    def get_cells(self, true_cells_only=True):
         """
         Get cells of the region.
 
         Raises ValueError if `true_cells_only` is True and the region kind does
-        not allow cells (e.g. surface integration region). For
-        `true_cells_only` equal to False, cells incident to facets are returned
-        if the region itself contains no cells.
-
-        If `offset` is True, the cell group offset is subtracted from the cell
-        ids.
+        not allow cells. For `true_cells_only` equal to False, cells incident
+        to facets are returned if the region itself contains no cells. Obeys
+        parent region, if given.
         """
-        cmesh = self.domain.cmesh
-
         if self.cells.shape[0] == 0:
             if true_cells_only:
                 msg = 'region %s has not true cells! (has kind: %s)' \
@@ -671,46 +596,76 @@ class Region(Struct):
 
             else:
                 # Has to be consistent with get_facet_indices()!
+                cmesh = self.domain.cmesh
                 cmesh.setup_connectivity(self.tdim - 1, self.tdim)
                 out = cmesh.get_incident(self.tdim, self.facets, self.tdim - 1)
 
-                igs = cmesh.cell_groups[out]
-                ic = nm.where(igs == ig)
-                out = out[ic]
+                if self.parent is not None:
+                    pcells = self.domain.regions[self.parent].cells
+                    ip = nm.in1d(out, pcells, assume_unique=False)
+                    out = out[ip]
 
         else:
-            out = cmesh.get_from_cell_group(ig, self.tdim, self.cells)
-
-        if offset:
-            out -= self.domain.cell_offsets[ig]
+            out = self.cells
 
         return out
 
-    def get_facet_indices(self, ig, offset=True, force_ig=True):
+    def get_cell_indices(self, cells, true_cells_only=True):
+        """
+        Return indices of `cells` in the region cells.
+
+        Raises ValueError if `true_cells_only` is True and the region kind does
+        not allow cells. For `true_cells_only` equal to False, cells incident
+        to facets are returned if the region itself contains no cells.
+
+        Notes
+        -----
+        If the number of unique values in `cells` is smaller or equal to the
+        number of cells in the region, all `cells` has to be also the region
+        cells (`self` is a superset of `cells`). The region cells are
+        considered depending on `true_cells_only`.
+
+        Otherwise, indices of all cells in `self` that are in `cells` are
+        returned.
+        """
+        fcells = self.get_cells(true_cells_only=true_cells_only)
+
+        if len(nm.unique(cells)) <= len(nm.unique(fcells)):
+            # self is a superset of cells.
+            ii = nm.searchsorted(fcells, cells)
+            assert_((fcells[ii] == cells).all())
+
+        else:
+            aux = nm.searchsorted(cells, fcells)
+            ii = nm.where(nm.take(cells, aux, mode='clip') == fcells)[0]
+
+        return ii
+
+    def get_facet_indices(self):
         """
         Return an array (per group) of (iel, ifa) for each facet. A facet can
         be in 1 (surface) or 2 (inner) cells.
-
-        If `offset` is True, the cell group offset is subtracted from the cell
-        ids.
-
-        If `force_ig` is True, only the cells with the given `ig` are used.
         """
         cmesh = self.domain.cmesh
-        facets = self.get_facets(ig)
+        cmesh.setup_connectivity(self.tdim - 1, self.tdim)
+
+        facets = self.facets
         cells, offs = cmesh.get_incident(self.tdim, facets, self.tdim - 1,
                                          ret_offsets=True)
+
+        if self.parent is not None:
+            pcells = self.domain.regions[self.parent].cells
+            ip = nm.in1d(cells, pcells, assume_unique=False)
+            cells = cells[ip]
+
+            counts = nm.diff(offs).astype(nm.int32)
+            pos = nm.repeat(nm.arange(facets.shape[0], dtype=nm.int32), counts)
+            new_counts = nm.bincount(pos, weights=ip).astype(nm.uint32)
+            offs = nm.cumsum(nm.r_[0, new_counts], dtype=nm.uint32)
+
         ii = cmesh.get_local_ids(facets, self.tdim - 1, cells, offs, self.tdim)
         fis = nm.c_[cells, ii]
 
-        if force_ig:
-            igs = cmesh.cell_groups[cells]
-            ic = nm.where(igs == ig)
-            fis = fis[ic]
-
-        if offset:
-            fis[:, 0] -= self.domain.cell_offsets[ig]
-
         return fis
 
     def setup_mirror_region(self):
@@ -719,51 +674,27 @@ class Region(Struct):
         """
         regions = self.domain.regions
 
-        parent = regions[self.parent]
         for reg in regions:
             mirror_parent = regions.find(reg.parent)
             if mirror_parent is None: continue
             if ((reg is not self)
-                and (len(reg.igs) == len(self.igs))
-                and (not len(nm.intersect1d(parent.igs, mirror_parent.igs)))
                 and nm.all(self.vertices == reg.vertices)):
                 mirror_region = reg
                 break
         else:
             raise ValueError('cannot find mirror region! (%s)' % self.name)
 
-        ig_map = {}
-        ig_map_i = {}
-        for igr in parent.igs:
-            v1 = self.get_vertices(igr)
-            for igc in mirror_parent.igs:
-                v2 = self.get_vertices(igc)
-                if nm.all(v1 == v2):
-                    ig_map[igc] = igr
-                    ig_map_i[igr] = igc
-                    break
-            else:
-                raise ValueError('cannot find mirror region group! (%d)' % igr)
-
-        self._igs = parent._igs
-        mirror_region._igs = mirror_parent._igs
-
         self.mirror_region = mirror_region
-        self.ig_map = ig_map
-        self.ig_map_i = ig_map_i
 
     def get_mirror_region(self):
-        return self.mirror_region, self.ig_map, self.ig_map_i
+        return self.mirror_region
 
-    def get_n_cells(self, ig=None, is_surface=False):
+    def get_n_cells(self, is_surface=False):
         """
         Get number of region cells.
 
         Parameters
         ----------
-        ig : int, optional
-            The group index. If None, counts from all groups are added
-            together.
         is_surface : bool
             If True, number of edges or faces according to domain
             dimension is returned instead.
@@ -773,31 +704,11 @@ class Region(Struct):
         n_cells : int
             The number of cells.
         """
-        if ig is not None:
-            if is_surface:
-                if self.domain.groups[ig].shape.dim == 2:
-                    return self.shape[ig].n_edge
-
-                else:
-                    return self.shape[ig].n_face
-
-            else:
-                return self.shape[ig].n_cell
+        if is_surface:
+            return self.shape.n_facet
 
         else:
-            return sum(self.get_n_cells(ig, is_surface=is_surface)
-                       for ig in self.igs)
-
-    def iter_cells(self):
-        ii = 0
-        off = 0
-        for ig in self.igs:
-            n_cell = self.shape[ig].n_cell
-            for iel in self.cells[off : off + n_cell]:
-                yield ig, ii, iel - off
-                ii += 1
-
-            off += n_cell
+            return self.shape.n_cell
 
     def has_cells(self):
         return self.cells.size > 0
@@ -816,14 +727,6 @@ class Region(Struct):
 
         return len(nm.intersect1d(se, oe))
 
-    def get_cell_offsets(self):
-        offs = {}
-        off = 0
-        for ig in self.igs:
-            offs[ig] = off
-            off += self.shape[ig].n_cell
-        return offs
-
     def get_charfun(self, by_cell=False, val_by_id=False):
         """
         Return the characteristic function of the region as a vector of values
@@ -858,7 +761,7 @@ class Region(Struct):
         cmesh = self.domain.cmesh
 
         e_verts = cmesh.get_incident(0, self.edges, 1)
-        e_verts.shape = (e_verts.shape[0] / 2, 2)
+        e_verts.shape = (e_verts.shape[0] // 2, 2)
 
         ii = nm.where(e_verts[:, 0] != e_verts[:, 1])[0]
         edges = self.edges[ii]
