@@ -1,6 +1,7 @@
 """
 Classes of equations composed of terms.
 """
+from __future__ import absolute_import
 import time
 from copy import copy
 
@@ -12,12 +13,13 @@ from sfepy.base.base import OneTypeList, Container, Struct
 from sfepy.discrete import Materials, Variables, create_adof_conns
 from sfepy.discrete.common.extmods.cmesh import create_mesh_graph
 from sfepy.terms import Terms, Term
+import six
 
 def parse_definition(equation_def):
     """
     Parse equation definition string to create term description list.
     """
-    from parse_equations import create_bnf
+    from .parse_equations import create_bnf
 
     term_descs = []
     bnf = create_bnf(term_descs)
@@ -56,7 +58,7 @@ class Equations(Container):
         conf = copy(conf)
 
         ii = 0
-        for name, desc in conf.iteritems():
+        for name, desc in six.iteritems(conf):
             if verbose:
                 output('equation "%s":' %  name)
                 output(desc)
@@ -81,6 +83,26 @@ class Equations(Container):
 
         self.collect_conn_info()
 
+    def add_equation(self, equation):
+        """
+        Add a new equation.
+
+        Parameters
+        ----------
+        equation : Equation instance
+                    The new equation.
+        """
+        self.append(equation)
+        self.variables.extend(
+             set(equation.collect_variables() ) - set(self.variables)
+        )
+        self.materials.extend(
+             set(equation.collect_materials() ) - set(self.materials)
+        )
+        equation.collect_conn_info(self.conn_info)
+        if not self.domain:
+           self.domain = self.get_domain()
+
     def create_subequations(self, var_names, known_var_names=None):
         """
         Create sub-equations containing only terms with the given virtual
@@ -235,9 +257,7 @@ class Equations(Container):
         for eq in self:
             output('  %s:' % eq.name)
             for term in eq.terms:
-                output('    %+.2e * %s.%d.%s(%s)'
-                       % (term.sign, term.name, term.integral.order,
-                          term.region.name, term.arg_str))
+                output('    %s' % term.get_str())
 
     def time_update(self, ts, ebcs=None, epbcs=None, lcbcs=None,
                     functions=None, problem=None, active_only=True,
@@ -283,7 +303,7 @@ class Equations(Container):
         active_bcs = self.variables.equation_mapping(ebcs, epbcs, ts, functions,
                                                      problem=problem,
                                                      active_only=active_only)
-        graph_changed = active_only and (active_bcs != self.active_bcs)
+        graph_changed = active_bcs != self.active_bcs
         self.active_bcs = active_bcs
 
         if graph_changed or not self.variables.adof_conns:
@@ -372,7 +392,7 @@ class Equations(Container):
             is_surface = rvar.is_surface or cvar.is_surface
 
             dct = info.dc_type.type
-            if not (dct in ('volume', 'scalar', 'plate') or is_surface
+            if not (dct in ('volume', 'scalar', 'custom') or is_surface
                     or info.is_trace or any_dof_conn):
                 continue
 
@@ -608,11 +628,15 @@ class Equations(Container):
             eqs = [self[eq] for eq in names]
 
         if mode == 'weak':
+            extras = []
             for eq in eqs:
-                eq.evaluate(mode=mode, dw_mode=dw_mode,
-                            term_mode=term_mode, asm_obj=asm_obj)
+                out = eq.evaluate(mode=mode, dw_mode=dw_mode,
+                                  term_mode=term_mode, asm_obj=asm_obj)
+                if isinstance(out, tuple): extras.extend(out[1])
 
             out = asm_obj
+            for extra in extras:
+                out = out + extra
 
         else:
             out = {}
@@ -723,17 +747,16 @@ class Equations(Container):
                 ic = get_indx(cname, stripped=True, allow_dual=True)
 
                 tangent_matrix.data[:] = 0.0
-                eq.evaluate(mode='weak', dw_mode='matrix',
-                            asm_obj=tangent_matrix)
+                aux = eq.evaluate(mode='weak', dw_mode='matrix',
+                                  asm_obj=tangent_matrix)
 
-                out[key] = tangent_matrix[ir, ic]
+                out[key] = aux[ir, ic]
 
         else:
             tangent_matrix.data[:] = 0.0
 
-            self.evaluate(mode='weak', dw_mode='matrix', asm_obj=tangent_matrix)
-
-            out = tangent_matrix
+            out = self.evaluate(mode='weak', dw_mode='matrix',
+                                asm_obj=tangent_matrix)
 
         return out
 
@@ -830,8 +853,11 @@ class Equation(Struct):
                                                       ret_status=True)
                     term.assemble_to(asm_obj, val, iels, mode=dw_mode)
 
+                out = asm_obj
+
             elif dw_mode == 'matrix':
 
+                extras = []
                 for term in self.terms:
                     svars = term.get_state_variables(unknown_only=True)
 
@@ -841,14 +867,15 @@ class Equation(Struct):
                                                           diff_var=svar.name,
                                                           standalone=False,
                                                           ret_status=True)
-                        term.assemble_to(asm_obj, val, iels,
-                                         mode=dw_mode, diff_var=svar)
+                        extra = term.assemble_to(asm_obj, val, iels,
+                                                 mode=dw_mode, diff_var=svar)
+                        if extra is not None: extras.append(extra)
+
+                out = (asm_obj, extras) if len(extras) else asm_obj
 
             else:
                 raise ValueError('unknown assembling mode! (%s)' % dw_mode)
 
-            out = asm_obj
-
         else:
             raise ValueError('unknown evaluation mode! (%s)' % mode)
 
