@@ -1,3 +1,4 @@
+from __future__ import absolute_import
 from copy import copy
 
 import numpy as nm
@@ -5,19 +6,41 @@ import numpy as nm
 from sfepy.base.base import output, get_default, OneTypeList, Struct, basestr
 from sfepy.discrete import Equations, Variables, Region, Integral, Integrals
 from sfepy.discrete.common.fields import setup_extra_data
+import six
+
+def apply_ebc_to_matrix(mtx, ebc_rows, epbc_rows=None):
+    """
+    Apply E(P)BC to matrix rows: put 1 to the diagonal for EBC DOFs, 1 to the
+    diagonal for master EPBC DOFs, -1 to the [master, slave] entries. It is
+    assumed, that the matrix contains zeros in EBC and master EPBC DOFs rows
+    and columns.
+    """
+    data, prows, cols = mtx.data, mtx.indptr, mtx.indices
+    # Does not change the sparsity pattern.
+    for ir in ebc_rows:
+        for ic in range(prows[ir], prows[ir + 1]):
+            if (cols[ic] == ir):
+                data[ic] = 1.0
+
+    if epbc_rows is not None:
+        master, slave = epbc_rows
+
+        # Changes sparsity pattern in-place - allocates new entries! The master
+        # DOFs are not allocated by Equations.create_matrix_graph(), see
+        # create_adof_conns().
+        mtx[master, master] = 1.0
+        mtx[master, slave] = -1.0
 
 ##
 # 02.10.2007, c
-class Evaluator( Struct ):
-    pass
-
-##
-# 02.10.2007, c
-class BasicEvaluator( Evaluator ):
+class Evaluator(Struct):
+    """
+    This class provides the functions required by a nonlinear solver for a
+    given problem.
+    """
 
     def __init__(self, problem, matrix_hook=None):
-        Evaluator.__init__(self, problem=problem,
-                           matrix_hook=matrix_hook)
+        Struct.__init__(self, problem=problem, matrix_hook=matrix_hook)
 
     def new_ulf_iteration(self, nls, vec, it, err, err0):
 
@@ -37,7 +60,8 @@ class BasicEvaluator( Evaluator ):
         nods = state.field.get_dofs_in_region(state.field.region, merge=True)
         coors = pb.domain.get_mesh_coors().copy()
         vs = state()
-        coors[nods,:] = coors[nods,:] + vs.reshape(len(nods), state.n_components)
+        coors[nods,:] = coors[nods,:] + vs.reshape(len(nods),
+                                                   state.n_components)
         if pb.ts.step == 1 and it == 0:
             state.field.save_mappings()
 
@@ -45,21 +69,30 @@ class BasicEvaluator( Evaluator ):
         pb.set_mesh_coors(coors, update_fields=False, actual=True,
                           clear_all=False)
 
-    def eval_residual( self, vec, is_full = False ):
-        if not is_full:
-            vec = self.make_full_vec( vec )
+    def eval_residual(self, vec, is_full=False):
+        if not is_full and self.problem.active_only:
+            vec = self.make_full_vec(vec)
 
         vec_r = self.problem.equations.eval_residuals(vec)
         if self.matrix_hook is not None:
             vec_r = self.matrix_hook(vec_r, self.problem, call_mode='residual')
 
+        if self.problem.equations.variables.has_lcbc:
+            mtx_lcbc = self.problem.equations.get_lcbc_operator()
+
+            vec_rr = mtx_lcbc.T * vec_r
+            if self.matrix_hook is not None:
+                vec_rr = self.matrix_hook(vec_rr, self.problem,
+                                          call_mode='lcbc_residual')
+            vec_r = vec_rr
+
         return vec_r
 
-    def eval_tangent_matrix( self, vec, mtx = None, is_full = False ):
+    def eval_tangent_matrix(self, vec, mtx=None, is_full=False):
         if isinstance(vec, basestr) and vec == 'linear':
             return get_default(mtx, self.problem.mtx_a)
 
-        if not is_full:
+        if not is_full and self.problem.active_only:
             vec = self.make_full_vec( vec )
 
         pb = self.problem
@@ -67,65 +100,35 @@ class BasicEvaluator( Evaluator ):
             mtx = pb.mtx_a
         mtx = pb.equations.eval_tangent_matrices(vec, mtx)
 
+        if not pb.active_only:
+            apply_ebc_to_matrix(mtx, *pb.get_ebc_indices())
+
         if self.matrix_hook is not None:
             mtx = self.matrix_hook(mtx, pb, call_mode='basic')
 
+        if self.problem.equations.variables.has_lcbc:
+            mtx_lcbc = self.problem.equations.get_lcbc_operator()
+
+            mtx_r = mtx_lcbc.T * mtx * mtx_lcbc
+            mtx_r = mtx_r.tocsr()
+            mtx_r.sort_indices()
+
+            if self.matrix_hook is not None:
+                mtx_r = self.matrix_hook(mtx_r, self.problem, call_mode='lcbc')
+
+            mtx = mtx_r
+
         return mtx
 
-    def make_full_vec( self, vec ):
+    def make_full_vec(self, vec):
         return self.problem.equations.make_full_vec(vec)
 
-##
-# 04.10.2007, c
-class LCBCEvaluator( BasicEvaluator ):
-
-    ##
-    # 04.10.2007, c
-    def __init__(self, problem, matrix_hook=None):
-        BasicEvaluator.__init__(self, problem, matrix_hook=matrix_hook)
-        self.mtx_lcbc = problem.equations.get_lcbc_operator()
-
-    ##
-    # 04.10.2007, c
-    def eval_residual( self, vec, is_full = False ):
-        if not is_full:
-            vec = self.make_full_vec( vec )
-        vec_r = BasicEvaluator.eval_residual( self, vec, is_full = True )
-        vec_rr = self.mtx_lcbc.T * vec_r
-        if self.matrix_hook is not None:
-            vec_rr = self.matrix_hook(vec_rr, self.problem,
-                                      call_mode='lcbc_residual')
-        return vec_rr
-
-    ##
-    # 04.10.2007, c
-    def eval_tangent_matrix( self, vec, mtx = None, is_full = False ):
-        if isinstance(vec, basestr) and vec == 'linear':
-            return get_default(mtx, self.problem.mtx_a)
-
-        if not is_full:
-            vec = self.make_full_vec( vec )
-        mtx = BasicEvaluator.eval_tangent_matrix( self, vec, mtx = mtx,
-                                                  is_full = True )
-        mtx_r = self.mtx_lcbc.T * mtx * self.mtx_lcbc
-        mtx_r = mtx_r.tocsr()
-        mtx_r.sort_indices()
-##         import pylab
-##         from sfepy.base.plotutils import spy
-##         spy( mtx_r )
-##         pylab.show()
-##         print mtx_r.__repr__()
-
-        if self.matrix_hook is not None:
-            mtx_r = self.matrix_hook(mtx_r, self.problem, call_mode='lcbc')
-
-        return mtx_r
-
 def create_evaluable(expression, fields, materials, variables, integrals,
                      regions=None,
-                     ebcs=None, epbcs=None, lcbcs=None, ts=None, functions=None,
+                     ebcs=None, epbcs=None, lcbcs=None,
+                     ts=None, functions=None,
                      auto_init=False, mode='eval', extra_args=None,
-                     verbose=True, kwargs=None):
+                     active_only=True, verbose=True, kwargs=None):
     """
     Create evaluable object (equations and corresponding variables)
     from the `expression` string.
@@ -168,6 +171,9 @@ def create_evaluable(expression, fields, materials, variables, integrals,
         each term region.
     extra_args : dict, optional
         Extra arguments to be passed to terms in the expression.
+    active_only : bool
+        If True, in 'weak' mode, the (tangent) matrices and residual
+        vectors (right-hand sides) contain only active DOFs.
     verbose : bool
         If False, reduce verbosity.
     kwargs : dict, optional
@@ -191,7 +197,7 @@ def create_evaluable(expression, fields, materials, variables, integrals,
         regions = OneTypeList(Region, regions)
 
     else:
-        regions = fields[fields.keys()[0]].domain.regions
+        regions = fields[list(fields.keys())[0]].domain.regions
 
     # Create temporary variables.
     aux_vars = Variables(variables)
@@ -219,9 +225,12 @@ def create_evaluable(expression, fields, materials, variables, integrals,
 
     if mode == 'weak':
         equations.time_update(ts, ebcs, epbcs, lcbcs, functions,
-                              verbose=verbose)
+                              active_only=active_only, verbose=verbose)
 
     else:
+        for eq in equations:
+            for term in eq.terms:
+                term.time_update(ts)
         setup_extra_data(equations.conn_info)
 
     return equations, variables
@@ -229,7 +238,7 @@ def create_evaluable(expression, fields, materials, variables, integrals,
 
 def eval_equations(equations, variables, names=None, preserve_caches=False,
                    mode='eval', dw_mode='vector', term_mode=None,
-                   verbose=True):
+                   active_only=True, verbose=True):
     """
     Evaluate the equations.
 
@@ -254,6 +263,9 @@ def eval_equations(equations, variables, names=None, preserve_caches=False,
         The term call mode - some terms support different call modes
         and depending on the call mode different values are
         returned.
+    active_only : bool
+        If True, in 'weak' mode, the (tangent) matrices and residual
+        vectors (right-hand sides) contain only active DOFs.
     verbose : bool
         If False, reduce verbosity.
 
@@ -271,7 +283,8 @@ def eval_equations(equations, variables, names=None, preserve_caches=False,
             asm_obj = equations.create_stripped_state_vector()
 
         else:
-            asm_obj = equations.create_matrix_graph(verbose=verbose)
+            asm_obj = equations.create_matrix_graph(active_only=active_only,
+                                                    verbose=verbose)
 
     if not preserve_caches:
         equations.invalidate_term_caches()
@@ -294,7 +307,8 @@ def eval_equations(equations, variables, names=None, preserve_caches=False,
 def eval_in_els_and_qp(expression, iels, coors,
                        fields, materials, variables,
                        functions=None, mode='eval', term_mode=None,
-                       extra_args=None, verbose=True, kwargs=None):
+                       extra_args=None, active_only=True, verbose=True,
+                       kwargs=None):
     """
     Evaluate an expression in given elements and points.
 
@@ -320,6 +334,9 @@ def eval_in_els_and_qp(expression, iels, coors,
         returned.
     extra_args : dict, optional
         Extra arguments to be passed to terms in the expression.
+    active_only : bool
+        If True, in 'weak' mode, the (tangent) matrices and residual
+        vectors (right-hand sides) contain only active DOFs.
     verbose : bool
         If False, reduce verbosity.
     kwargs : dict, optional
@@ -334,39 +351,40 @@ def eval_in_els_and_qp(expression, iels, coors,
     weights = nm.ones_like(coors[:, 0])
     integral = Integral('ie', coors=coors, weights=weights)
 
-    domain = fields.values()[0].domain
+    domain = list(fields.values())[0].domain
 
     region = Region('Elements', 'given elements', domain, '')
     region.cells = iels
     region.update_shape()
     domain.regions.append(region)
 
-    for field in fields.itervalues():
+    for field in six.itervalues(fields):
         field.clear_mappings(clear_all=True)
         field.clear_qp_base()
 
     aux = create_evaluable(expression, fields, materials,
                            variables.itervalues(), Integrals([integral]),
-                           functions=functions,
-                           mode=mode, extra_args=extra_args, verbose=verbose,
-                           kwargs=kwargs)
+                           functions=functions, mode=mode,
+                           extra_args=extra_args, active_only=active_only,
+                           verbose=verbose, kwargs=kwargs)
     equations, variables = aux
 
     out = eval_equations(equations, variables,
                          preserve_caches=False,
-                         mode=mode, term_mode=term_mode)
+                         mode=mode, term_mode=term_mode,
+                         active_only=active_only)
     domain.regions.pop()
 
     return out
 
 def assemble_by_blocks(conf_equations, problem, ebcs=None, epbcs=None,
-                       dw_mode='matrix'):
+                       dw_mode='matrix', active_only=True):
     """Instead of a global matrix, return its building blocks as defined in
     `conf_equations`. The name and row/column variables of each block have to
     be encoded in the equation's name, as in::
 
         conf_equations = {
-          'A,v,u' : "dw_lin_elastic_iso.i1.Y2( inclusion.lame, v, u )",
+          'A,v,u' : "dw_lin_elastic.i1.Y2( inclusion.D, v, u )",
         }
 
     Notes
@@ -382,7 +400,7 @@ def assemble_by_blocks(conf_equations, problem, ebcs=None, epbcs=None,
         raise TypeError('bad BC!')
 
     matrices = {}
-    for key, mtx_term in conf_equations.iteritems():
+    for key, mtx_term in six.iteritems(conf_equations):
         ks = key.split( ',' )
         mtx_name, var_names = ks[0], ks[1:]
         output( mtx_name, var_names )
@@ -403,7 +421,7 @@ def assemble_by_blocks(conf_equations, problem, ebcs=None, epbcs=None,
         problem.update_materials()
         mtx = problem.evaluate(mtx_term, auto_init=True,
                                mode='weak', dw_mode='matrix',
-                               copy_materials=False)
+                               copy_materials=False, active_only=active_only)
         matrices[mtx_name] = mtx[ir,ic]
 
     return matrices
