@@ -11,6 +11,7 @@ Important attributes of continuous (order > 0) :class:`Field` and
 where `conn` is the mesh vertex connectivity, `econn` is the
 region-local field connectivity.
 """
+from __future__ import absolute_import
 import numpy as nm
 
 from sfepy.base.base import output, get_default, assert_
@@ -26,17 +27,20 @@ from sfepy.discrete.fem.fe_surface import FESurface
 from sfepy.discrete.integrals import Integral
 from sfepy.discrete.fem.linearizer import (get_eval_dofs, get_eval_coors,
                                            create_output)
+import six
 
 def set_mesh_coors(domain, fields, coors, update_fields=False, actual=False,
-                   clear_all=True):
+                   clear_all=True, extra_dofs=False):
     if actual:
-        domain.mesh.coors_act = coors.copy()
+        if not hasattr(domain.mesh, 'coors_act'):
+            domain.mesh.coors_act = nm.zeros_like(domain.mesh.coors)
+        domain.mesh.coors_act[:] = coors[:domain.mesh.n_nod]
     else:
-        domain.cmesh.coors[:] = coors
+        domain.cmesh.coors[:] = coors[:domain.mesh.n_nod]
 
     if update_fields:
-        for field in fields.itervalues():
-            field.setup_coors(coors)
+        for field in six.itervalues(fields):
+            field.set_coors(coors, extra_dofs=extra_dofs)
             field.clear_mappings(clear_all=clear_all)
 
 def eval_nodal_coors(coors, mesh_coors, region, poly_space, geom_poly_space,
@@ -252,6 +256,10 @@ class FEField(Field):
         self.setup_coors()
         self.clear_mappings(clear_all=True)
         self.clear_qp_base()
+        self.basis_transform = None
+        self.econn0 = None
+        self.unused_dofs = None
+        self.stored_subs = None
 
     def _set_approx_order(self, approx_order):
         """
@@ -347,16 +355,20 @@ class FEField(Field):
                 efs = efs[:,nm.newaxis]
             self.efaces = nm.hstack((self.efaces, efs))
 
-    def setup_coors(self, coors=None):
-        """
-        Setup coordinates of field nodes.
-        """
-        mesh = self.domain.mesh
-        self.coors = nm.empty((self.n_nod, mesh.dim), nm.float64)
+        if gel.dim == 3:
+            self.eedges = gel.edges.copy()
+            efs = node_desc.edge
+            if efs is not None:
+                efs = nm.array(efs).squeeze()
 
-        if coors is None:
-            coors = mesh.coors
+                if efs.ndim < 2:
+                    efs = efs[:,nm.newaxis]
+                self.eedges = nm.hstack((self.eedges, efs))
 
+    def set_coors(self, coors, extra_dofs=False):
+        """
+        Set coordinates of field nodes.
+        """
         # Mesh vertex nodes.
         if self.n_vertex_dof:
             indx = self.vertex_remap_i
@@ -364,10 +376,27 @@ class FEField(Field):
                                                      indx.astype(nm.int32),
                                                      axis=0)
 
-        gps = self.gel.poly_space
-        ps = self.poly_space
+        n_ex_dof = self.n_bubble_dof + self.n_edge_dof + self.n_face_dof
 
-        eval_nodal_coors(self.coors, coors, self.region, ps, gps, self.econn)
+        # extra nodes
+        if n_ex_dof:
+            if extra_dofs:
+                if self.n_nod != coors.shape[0]:
+                    raise NotImplementedError
+                self.coors[:] = coors
+            else:
+                gps = self.gel.poly_space
+                ps = self.poly_space
+                eval_nodal_coors(self.coors, coors, self.region,
+                                 ps, gps, self.econn)
+
+    def setup_coors(self):
+        """
+        Setup coordinates of field nodes.
+        """
+        mesh = self.domain.mesh
+        self.coors = nm.empty((self.n_nod, mesh.dim), nm.float64)
+        self.set_coors(mesh.coors)
 
     def get_vertices(self):
         """
@@ -388,11 +417,10 @@ class FEField(Field):
         ----------
         integral : Integral instance
             The integral describing used numerical quadrature.
-        integration : 'volume', 'plate', 'surface', 'surface_extra' or 'point'
+        integration : 'volume', 'surface', 'surface_extra', 'point' or 'custom'
             The term integration type.
         region_name : str
-            The name of surface region, required when `shape_kind` is
-            'surface'.
+            The name of the region of the integral.
 
         Returns
         -------
@@ -427,7 +455,7 @@ class FEField(Field):
             else:
                 data_shape = (sd.n_fa, n_qp, dim, self.econn.shape[1])
 
-        elif integration in ('volume', 'plate'):
+        elif integration in ('volume', 'custom'):
             _, weights = integral.get_qp(self.gel.name)
             n_qp = weights.shape[0]
 
@@ -497,7 +525,7 @@ class FEField(Field):
         """
         qpkey = (integral.order, key)
 
-        if not self.qp_coors.has_key(qpkey):
+        if qpkey not in self.qp_coors:
             if (key[0] == 's') and not self.is_surface:
                 dim = self.gel.dim - 1
                 n_fp = self.gel.surface_facet.n_vertex
@@ -511,6 +539,83 @@ class FEField(Field):
 
         return self.qp_coors[qpkey]
 
+    def substitute_dofs(self, subs, restore=False):
+        """
+        Perform facet DOF substitutions according to `subs`.
+
+        Modifies `self.econn` in-place and sets `self.econn0`,
+        `self.unused_dofs` and `self.basis_transform`.
+        """
+        if restore and (self.stored_subs is not None):
+            self.econn0 = self.econn
+            self.econn, self.unused_dofs, basis_transform = self.stored_subs
+
+        else:
+            if subs is None:
+                self.econn0 = self.econn
+                return
+
+            else:
+                self.econn0 = self.econn.copy()
+
+            self._substitute_dofs(subs)
+
+            self.unused_dofs = nm.setdiff1d(self.econn0, self.econn)
+
+            basis_transform = self._eval_basis_transform(subs)
+
+        self.set_basis_transform(basis_transform)
+
+    def restore_dofs(self, store=False):
+        """
+        Undoes the effect of :func:`FEField.substitute_dofs()`.
+        """
+        if self.econn0 is None:
+            raise ValueError('no original DOFs to restore!')
+
+        if store:
+            self.stored_subs = (self.econn,
+                                self.unused_dofs,
+                                self.basis_transform)
+
+        else:
+            self.stored_subs = None
+
+        self.econn = self.econn0
+        self.econn0 = None
+        self.unused_dofs = None
+        self.basis_transform = None
+
+    def set_basis_transform(self, transform):
+        """
+        Set local element basis transformation.
+
+        The basis transformation is applied in :func:`FEField.get_base()` and
+        :func:`FEField.create_mapping()`.
+
+        Parameters
+        ----------
+        transform : array, shape `(n_cell, n_ep, n_ep)`
+            The array with `(n_ep, n_ep)` transformation matrices for each cell
+            in the field's region, where `n_ep` is the number of element DOFs.
+        """
+        self.basis_transform = transform
+
+    def restore_substituted(self, vec):
+        """
+        Restore values of the unused DOFs using the transpose of the applied
+        basis transformation.
+        """
+        if (self.econn0 is None) or (self.basis_transform is None):
+            raise ValueError('no original DOF values to restore!!')
+
+        vec = vec.reshape((self.n_nod, self.n_components)).copy()
+        evec = vec[self.econn]
+
+        vec[self.econn0] = nm.einsum('cji,cjk->cik', self.basis_transform, evec)
+
+        return vec.ravel()
+
     def get_base(self, key, derivative, integral, iels=None,
                  from_geometry=False, base_only=True):
         qp = self.get_qp(key, integral)
@@ -524,14 +629,15 @@ class FEField(Field):
         _key = key if not from_geometry else 'g' + key
         bf_key = (integral.order, _key, derivative)
 
-        if not self.bf.has_key(bf_key):
+        if bf_key not in self.bf:
             if (iels is not None) and (self.ori is not None):
                 ori = self.ori[iels]
 
             else:
                 ori = self.ori
 
-            self.bf[bf_key] = ps.eval_base(qp.vals, diff=derivative, ori=ori)
+            self.bf[bf_key] = ps.eval_base(qp.vals, diff=derivative, ori=ori,
+                                           transform=self.basis_transform)
 
         if base_only:
             return self.bf[bf_key]
@@ -693,8 +799,7 @@ class FEField(Field):
                               data=dofs, var_name=var_name,
                               dofs=dof_names, field_name=self.name)
 
-        elif ((not self.is_higher_order())
-            or (linearization.kind == 'strip')):
+        elif linearization.kind == 'strip':
             if extend:
                 ext = self.extend_dofs(dofs, fill_value)
 
@@ -903,36 +1008,11 @@ class FEField(Field):
             conn = nm.take(dconn, iels.astype(nm.int32), axis=0)
             mapping = VolumeMapping(coors, conn, poly_space=geo_ps)
             vg = mapping.get_mapping(qp.vals, qp.weights, poly_space=ps,
-                                     ori=self.ori)
+                                     ori=self.ori,
+                                     transform=self.basis_transform)
 
             out = vg
 
-        elif integration == 'plate':
-            import sfepy.mechanics.membranes as mm
-            from sfepy.linalg import dot_sequences
-
-            qp = self.get_qp('v', integral)
-            iels = region.get_cells()
-
-            ps = self.interp.poly_spaces['v']
-            bf = self.get_base('v', 0, integral, iels=iels)
-
-            conn = nm.take(dconn, nm.int32(iels), axis=0)
-            ccoors = coors[conn]
-
-            # Coordinate transformation matrix (transposed!).
-            mtx_t = mm.create_transformation_matrix(ccoors)
-
-            # Transform coordinates to the local coordinate system.
-            coors_loc = dot_sequences((ccoors - ccoors[:, 0:1, :]), mtx_t)
-
-            # Mapping from transformed elements to reference elements.
-            mapping = mm.create_mapping(coors_loc, self.gel, 1)
-            vg = mapping.get_mapping(qp.vals, qp.weights, poly_space=ps,
-                                     ori=self.ori)
-            vg.mtx_t = mtx_t
-            out = vg
-
         elif (integration == 'surface') or (integration == 'surface_extra'):
             assert_(self.approx_order > 0)
 
@@ -954,7 +1034,7 @@ class FEField(Field):
                 self.create_bqp(region.name, integral)
                 qp = self.qp_coors[(integral.order, esd.bkey)]
 
-                abf = ps.eval_base(qp.vals[0])
+                abf = ps.eval_base(qp.vals[0], transform=self.basis_transform)
                 bf = abf[..., self.efaces[0]]
 
                 indx = self.gel.get_surface_entities()[0]
@@ -977,7 +1057,7 @@ class FEField(Field):
             else:
                 # Do not use BQP for surface fields.
                 qp = self.get_qp(sd.face_type, integral)
-                bf = ps.eval_base(qp.vals)
+                bf = ps.eval_base(qp.vals, transform=self.basis_transform)
 
                 sg = mapping.get_mapping(qp.vals, qp.weights,
                                          poly_space=Struct(n_nod=bf.shape[-1]),
@@ -988,8 +1068,11 @@ class FEField(Field):
         elif integration == 'point':
             out = mapping = None
 
+        elif integration == 'custom':
+            raise ValueError('cannot create custom mapping!')
+
         else:
-            raise ValueError('unknown inegration geometry type: %s'
+            raise ValueError('unknown integration geometry type: %s'
                              % integration)
 
         if out is not None:
@@ -1040,7 +1123,7 @@ class VolumeField(FEField):
         Setup the field region geometry.
         """
         cmesh = self.domain.cmesh
-        for key, gel in self.domain.geom_els.iteritems():
+        for key, gel in six.iteritems(self.domain.geom_els):
             ct = cmesh.cell_types
             if (ct[self.region.cells] == cmesh.key_to_index[gel.name]).all():
                 self.gel = gel
@@ -1117,7 +1200,7 @@ class VolumeField(FEField):
         elif dct == 'point':
             self.setup_point_data(self, info.region)
 
-        elif dct not in ('volume', 'scalar', 'plate'):
+        elif dct not in ('volume', 'scalar', 'custom'):
             raise ValueError('unknown dof connectivity type! (%s)' % dct)
 
     def setup_point_data(self, field, region):
@@ -1146,12 +1229,12 @@ class VolumeField(FEField):
         """
         ct = conn_type.type if isinstance(conn_type, Struct) else conn_type
 
-        if ct in ('volume', 'plate'):
+        if ct in ('volume', 'custom'):
             if region.name == self.region.name:
                 conn = self.econn
 
             else:
-                tco = integration in ('volume', 'plate')
+                tco = integration in ('volume', 'custom')
                 cells = region.get_cells(true_cells_only=tco)
                 ii = self.region.get_cell_indices(cells, true_cells_only=tco)
                 conn = nm.take(self.econn, ii, axis=0)
@@ -1231,16 +1314,24 @@ class SurfaceField(FEField):
         ok : bool
             True if the region is usable for the field.
         """
-        ok = ((region.kind_tdim == (region.tdim - 1))
-              and (region.get_n_cells(True) > 0))
+        ok1 = ((region.kind_tdim == (region.tdim - 1))
+               and (region.get_n_cells(True) > 0))
+        if not ok1:
+            output('bad region topological dimension and kind! (%d, %s)'
+                   % (region.tdim, region.kind))
 
-        return ok
+        n_ns = region.get_facet_indices().shape[0] - region.get_n_cells(True)
+        ok2 = n_ns == 0
+        if not ok2:
+            output('%d region facets are not on the domain surface!' % n_ns)
+
+        return ok1 and ok2
 
     def _setup_geometry(self):
         """
         Setup the field region geometry.
         """
-        for key, vgel in self.domain.geom_els.iteritems():
+        for key, vgel in six.iteritems(self.domain.geom_els):
             self.gel = vgel.surface_facet
             break
 
