@@ -1,8 +1,11 @@
+from __future__ import absolute_import
 import numpy as nm
 import numpy.linalg as nla
 
 from sfepy.base.base import find_subclasses, assert_, Struct
 from sfepy.linalg import combine, insert_strided_axis
+from six.moves import range
+from functools import reduce
 
 # Requires fixed vertex numbering!
 vertex_maps = {3 : [[0, 0, 0],
@@ -20,6 +23,26 @@ vertex_maps = {3 : [[0, 0, 0],
                1 : [[0],
                     [1]]}
 
+def transform_basis(transform, bf):
+    """
+    Transform a basis `bf` using `transform` array of matrices.
+    """
+    if bf.ndim == 3:
+        nbf = nm.einsum('cij,qdj->cqdi', transform, bf, order='C')
+
+    elif bf.ndim == 4:
+        if bf.shape[0] == 1:
+            nbf = nm.einsum('cij,qdj->cqdi', transform, bf[0], order='C')
+
+        else:
+            nbf = nm.einsum('cij,cqdj->cqdi', transform, bf, order='C')
+
+    # Note: the 2nd derivatives are not supported here.
+    # Workaround for NumPy 1.14.0 - order is ignored(?)
+    nbf = nm.ascontiguousarray(nbf)
+
+    return nbf
+
 class LagrangeNodes(Struct):
     """Helper class for defining nodes of Lagrange elements."""
 
@@ -275,23 +298,29 @@ class PolySpace(Struct):
 
         self.bbox = nm.vstack((geometry.coors.min(0), geometry.coors.max(0)))
 
-    def eval_base(self, coors, diff=False, ori=None, force_axis=False,
-                  suppress_errors=False, eps=1e-15):
+    def eval_base(self, coors, diff=0, ori=None, force_axis=False,
+                  transform=None, suppress_errors=False, eps=1e-15):
         """
-        Evaluate the basis in points given by coordinates. The real work is
-        done in _eval_base() implemented in subclasses.
+        Evaluate the basis or its first or second derivatives in points given
+        by coordinates. The real work is done in _eval_base() implemented in
+        subclasses.
+
+        Note that the second derivative code is a work-in-progress and only
+        `coors` and `transform` arguments are used.
 
         Parameters
         ----------
         coors : array_like
             The coordinates of points where the basis is evaluated. See Notes.
-        diff : bool
-            If True, return the first derivative.
+        diff : 0, 1 or 2
+            If nonzero, return the given derivative.
         ori : array_like, optional
             Optional orientation of element facets for per element basis.
         force_axis : bool
             If True, force the resulting array shape to have one more axis even
             when `ori` is None.
+        transform : array_like, optional
+            The basis transform array.
         suppress_errors : bool
             If True, do not report points outside the reference domain.
         eps : float
@@ -300,10 +329,11 @@ class PolySpace(Struct):
         Returns
         -------
         base : array
-            The basis (shape (n_coor, 1, n_base)) or its derivative (shape
-            (n_coor, dim, n_base)) evaluated in the given points. An additional
-            axis is pre-pended of length n_cell, if `ori` is given, or of
-            length 1, if `force_axis` is True.
+            The basis (shape (n_coor, 1, n_base)) or its first derivative
+            (shape (n_coor, dim, n_base)) or its second derivative (shape
+            (n_coor, dim, dim, n_base)) evaluated in the given points. An
+            additional axis is pre-pended of length n_cell, if `ori` is given,
+            or of length 1, if `force_axis` is True.
 
         Notes
         -----
@@ -343,6 +373,9 @@ class PolySpace(Struct):
                                            suppress_errors=suppress_errors,
                                            eps=eps)
 
+        if transform is not None:
+            base = transform_basis(transform, base)
+
         return base
 
     def get_mtx_i(self):
@@ -351,11 +384,62 @@ class PolySpace(Struct):
     def describe_nodes(self):
         return NodeDescription(self.nts, self.nodes)
 
-class LagrangeSimplexPolySpace(PolySpace):
+class LagrangePolySpace(PolySpace):
+
+    def create_context(self, cmesh, eps, check_errors, i_max, newton_eps,
+                       tdim=None):
+        from sfepy.discrete.fem.extmods.bases import CLagrangeContext
+
+        ref_coors = self.geometry.coors
+
+        if cmesh is not None:
+            mesh_coors = cmesh.coors
+
+            conn = cmesh.get_conn(cmesh.tdim, 0)
+            mesh_conn = conn.indices.reshape(cmesh.n_el, -1).astype(nm.int32)
+
+            if tdim is None:
+                tdim = cmesh.tdim
+
+        else:
+            mesh_coors = mesh_conn = None
+
+        if tdim is None:
+            raise ValueError('supply either cmesh or tdim!')
+
+        ctx = CLagrangeContext(order=self.order,
+                               tdim=tdim,
+                               nodes=self.nodes,
+                               ref_coors=ref_coors,
+                               mesh_coors=mesh_coors,
+                               mesh_conn=mesh_conn,
+                               mtx_i=self.get_mtx_i(),
+                               eps=eps,
+                               check_errors=check_errors,
+                               i_max=i_max,
+                               newton_eps=newton_eps)
+
+        return ctx
+
+    def _eval_base(self, coors, diff=0, ori=None,
+                   suppress_errors=False, eps=1e-15):
+        """
+        See :func:`PolySpace.eval_base()`.
+        """
+        if diff == 2:
+            base = self._eval_hessian(coors)
+
+        else:
+            base = self.eval_ctx.evaluate(coors, diff=diff,
+                                          eps=eps,
+                                          check_errors=not suppress_errors)
+        return base
+
+class LagrangeSimplexPolySpace(LagrangePolySpace):
     """Lagrange polynomial space on a simplex domain."""
     name = 'lagrange_simplex'
 
-    def __init__(self, name, geometry, order):
+    def __init__(self, name, geometry, order, init_context=True):
         PolySpace.__init__(self, name, geometry, order)
 
         n_v = geometry.n_vertex
@@ -369,15 +453,22 @@ class LagrangeSimplexPolySpace(PolySpace):
         self.node_coors = nm.ascontiguousarray(node_coors)
         self.n_nod = self.nodes.shape[0]
 
+        if init_context:
+            self.eval_ctx = self.create_context(None, 0, 1e-15, 100, 1e-8,
+                                                tdim=n_v - 1)
+
+        else:
+            self.eval_ctx = None
+
     def _define_nodes(self):
         # Factorial.
-        fac = lambda n : reduce(lambda a, b : a * (b + 1), xrange(n), 1)
+        fac = lambda n : reduce(lambda a, b : a * (b + 1), range(n), 1)
 
         geometry = self.geometry
         n_v, dim = geometry.n_vertex, geometry.dim
         order = self.order
 
-        n_nod = fac(order + dim) / (fac(order) * fac(dim))
+        n_nod = fac(order + dim) // (fac(order) * fac(dim))
         ## print n_nod, gd
         nodes = nm.zeros((n_nod, n_v), nm.int32)
         nts = nm.zeros((n_nod, 2), nm.int32)
@@ -432,24 +523,80 @@ class LagrangeSimplexPolySpace(PolySpace):
 
         return nodes, nts, node_coors
 
-    def _eval_base(self, coors, diff=False, ori=None,
-                   suppress_errors=False, eps=1e-15):
-        """See PolySpace.eval_base()."""
-        from extmods.bases import eval_lagrange_simplex
+    def _eval_hessian(self, coors):
+        """
+        Evaluate the second derivatives of the basis.
+        """
+        def get_bc(coor):
+            rhs = nm.concatenate((coor, [1]))
+            bc = nm.dot(self.mtx_i, rhs)
 
-        base = eval_lagrange_simplex(coors, self.mtx_i, self.nodes,
-                                     self.order, diff,
-                                     eps, not suppress_errors)
+            return bc
 
-        return base
+        def get_val(bc, node, omit=[]):
+            val = nm.ones(1, nm.float64)
+            for i1 in range(bc.shape[0]):
+                if i1 in omit: continue
+
+                for i2 in range(node[i1]):
+                    val *= (self.order * bc[i1] - i2) / (i2 + 1.0)
+
+            return val
+
+        def get_der(bc1, node1, omit=[]):
+            val = nm.zeros(1, nm.float64)
+            for i1 in range(node1):
+                if i1 in omit: continue
+
+                aux = nm.ones(1, nm.float64)
+                for i2 in range(node1):
+                    if (i1 == i2) or (i2 in omit): continue
+                    aux *= (self.order * bc1 - i2) / (i2 + 1.0)
+
+                val += aux * self.order / (i1 + 1.0)
+
+            return val
+
+        n_v = self.mtx_i.shape[0]
+        dim = n_v - 1
+
+        mi = self.mtx_i[:, :dim]
+        bfgg = nm.zeros((coors.shape[0], dim, dim, self.n_nod),
+                        dtype=nm.float64)
+
+        for ic, coor in enumerate(coors):
+            bc = get_bc(coor)
+
+            for ii, node in enumerate(self.nodes):
+                for ig1, bc1 in enumerate(bc): # 1. derivative w.r.t. bc1.
+                    for ig2, bc2 in enumerate(bc): # 2. derivative w.r.t. bc2.
+                        if ig1 == ig2:
+                            val = get_val(bc, node, omit=[ig1])
+
+                            vv = 0.0
+                            for i1 in range(node[ig1]):
+                                aux = get_der(bc2, node[ig2], omit=[i1])
+                                vv += aux * self.order / (i1 + 1.0)
+
+                            val *= vv
+
+                        else:
+                            val = get_val(bc, node, omit=[ig1, ig2])
+                            val *= get_der(bc1, node[ig1])
+                            val *= get_der(bc2, node[ig2])
+
+                        bfgg[ic, :, :, ii] += val * mi[ig1] * mi[ig2][:, None]
+
+        return bfgg
 
 class LagrangeSimplexBPolySpace(LagrangeSimplexPolySpace):
     """Lagrange polynomial space with forced bubble function on a simplex
     domain."""
     name = 'lagrange_simplex_bubble'
 
-    def __init__(self, name, geometry, order):
-        LagrangeSimplexPolySpace.__init__(self, name, geometry, order)
+    def __init__(self, name, geometry, order, init_context=True):
+        LagrangeSimplexPolySpace.__init__(self, name, geometry, order,
+                                          init_context=False)
 
         nodes, nts, node_coors = self.nodes, self.nts, self.node_coors
 
@@ -475,39 +622,44 @@ class LagrangeSimplexBPolySpace(LagrangeSimplexPolySpace):
 
         self.n_nod = self.nodes.shape[0]
 
-    def _eval_base(self, coors, diff=False, ori=None,
-                   suppress_errors=False, eps=1e-15):
-        """See PolySpace.eval_base()."""
-        from extmods.bases import eval_lagrange_simplex
+        if init_context:
+            self.eval_ctx = self.create_context(None, 0, 1e-15, 100, 1e-8,
+                                                tdim=n_v - 1)
 
-        base = eval_lagrange_simplex(coors, self.mtx_i, self.nodes[:-1],
-                                     self.order, diff,
-                                     eps, not suppress_errors)
-        bubble = eval_lagrange_simplex(coors, self.mtx_i, self.bnode,
-                                       int(self.bnode.sum()), diff,
-                                       eps, not suppress_errors)
+        else:
+            self.eval_ctx = None
 
-        base -= bubble / (self.n_nod - 1)
-        base = nm.ascontiguousarray(nm.dstack((base, bubble)))
+    def create_context(self, *args, **kwargs):
+        ctx = LagrangePolySpace.create_context(self, *args, **kwargs)
+        ctx.is_bubble = 1
 
-        return base
+        return ctx
 
-class LagrangeTensorProductPolySpace(PolySpace):
+class LagrangeTensorProductPolySpace(LagrangePolySpace):
     """Lagrange polynomial space on a tensor product domain."""
     name = 'lagrange_tensor_product'
 
-    def __init__(self, name, geometry, order):
+    def __init__(self, name, geometry, order, init_context=True):
         PolySpace.__init__(self, name, geometry, order)
 
         g1d = Struct(n_vertex = 2,
                      dim = 1,
-                     coors = self.bbox[:,0:1])
-        self.ps1d = LagrangeSimplexPolySpace('P_aux', g1d, order)
+                     coors = self.bbox[:,0:1].copy())
+        self.ps1d = LagrangeSimplexPolySpace('P_aux', g1d, order,
+                                             init_context=False)
 
         self.nodes, self.nts, node_coors = self._define_nodes()
         self.node_coors = nm.ascontiguousarray(node_coors)
         self.n_nod = self.nodes.shape[0]
 
+        if init_context:
+            tdim = int(nm.sqrt(geometry.n_vertex))
+            self.eval_ctx = self.create_context(None, 0, 1e-15, 100, 1e-8,
+                                                tdim=tdim)
+
+        else:
+            self.eval_ctx = None
+
     def _define_nodes(self):
         geometry = self.geometry
         order = self.order
@@ -597,11 +749,11 @@ class LagrangeTensorProductPolySpace(PolySpace):
         if diff:
             base = nm.ones((coors.shape[0], dim, self.n_nod), dtype=nm.float64)
 
-            for ii in xrange(dim):
+            for ii in range(dim):
                 self.ps1d.nodes = self.nodes[:,2*ii:2*ii+2].copy()
                 self.ps1d.n_nod = self.n_nod
 
-                for iv in xrange(dim):
+                for iv in range(dim):
                     if ii == iv:
                         base[:,iv:iv+1,:] *= ev(coors[:,ii:ii+1].copy(),
                                                 diff=True,
@@ -617,7 +769,7 @@ class LagrangeTensorProductPolySpace(PolySpace):
         else:
             base = nm.ones((coors.shape[0], 1, self.n_nod), dtype=nm.float64)
 
-            for ii in xrange(dim):
+            for ii in range(dim):
                 self.ps1d.nodes = self.nodes[:,2*ii:2*ii+2].copy()
                 self.ps1d.n_nod = self.n_nod
                 
@@ -628,16 +780,47 @@ class LagrangeTensorProductPolySpace(PolySpace):
 
         return base
 
-    def _eval_base(self, coors, diff=False, ori=None,
-                   suppress_errors=False, eps=1e-15):
-        """See PolySpace.eval_base()."""
-        from extmods.bases import eval_lagrange_tensor_product as ev
-        ps1d = self.ps1d
+    def _eval_hessian(self, coors):
+        """
+        Evaluate the second derivatives of the basis.
+        """
+        evh = self.ps1d.eval_base
 
-        base = ev(coors, ps1d.mtx_i, self.nodes, self.order, diff,
-                  eps, not suppress_errors)
+        dim = self.geometry.dim
+        bfgg = nm.zeros((coors.shape[0], dim, dim, self.n_nod),
+                        dtype=nm.float64)
 
-        return base
+        v0s = []
+        v1s = []
+        v2s = []
+        for ii in range(dim):
+            self.ps1d.nodes = self.nodes[:,2*ii:2*ii+2].copy()
+            self.ps1d.n_nod = self.n_nod
+            ev = self.ps1d.create_context(None, 0, 1e-15, 100, 1e-8,
+                                          tdim=1).evaluate
+
+            v0s.append(ev(coors[:, ii:ii+1].copy())[:, 0, :])
+            v1s.append(ev(coors[:, ii:ii+1].copy(), diff=1)[:, 0, :])
+            v2s.append(evh(coors[:, ii:ii+1], diff=2)[:, 0, 0, :])
+
+        for ir in range(dim):
+            vv = v2s[ir] # Destroys v2s!
+            for ik in range(dim):
+                if ik == ir: continue
+                vv *= v0s[ik]
+
+            bfgg[:, ir, ir, :] = vv
+
+            for ic in range(dim):
+                if ic == ir: continue
+                val = v1s[ir] * v1s[ic]
+                for ik in range(dim):
+                    if (ik == ir) or (ik == ic): continue
+                    val *= v0s[ik]
+
+                bfgg[:, ir, ic, :] += val
+
+        return bfgg
 
     def get_mtx_i(self):
         return self.ps1d.mtx_i
@@ -796,7 +979,7 @@ class LobattoTensorProductPolySpace(PolySpace):
         """
         See PolySpace.eval_base().
         """
-        from extmods.lobatto_bases import eval_lobatto_tensor_product as ev
+        from .extmods.lobatto_bases import eval_lobatto_tensor_product as ev
         c_min, c_max = self.bbox[:, 0]
 
         base = ev(coors, self.nodes, c_min, c_max, self.order, diff)
