@@ -1,9 +1,11 @@
 """Classes for probing values of Variables, for example, along a line."""
+from __future__ import absolute_import
 import numpy as nm
 import numpy.linalg as nla
 
 from sfepy.base.base import get_default, basestr, Struct
 from sfepy.linalg import make_axis_rotation_matrix, norm_l2_along_axis
+import six
 
 def write_results(filename, probe, results):
     """
@@ -22,7 +24,7 @@ def write_results(filename, probe, results):
     fd = open(filename, 'w') if isinstance(filename, basestr) else filename
 
     fd.write('\n'.join(probe.report()) + '\n')
-    for key, result in results.iteritems():
+    for key, result in six.iteritems(results):
         pars, vals = result
         fd.write('\n# %s %d\n' % (key, vals.shape[-1]))
 
@@ -160,6 +162,13 @@ class Probe(Struct):
 
         self.is_refined = False
 
+    def get_evaluate_cache(self):
+        """
+        Return the evaluate cache for domain-related data given by
+        `self.share_geometry`.
+        """
+        return Probe.cache if self.share_geometry else self.cache
+
     def set_n_point(self, n_point):
         """
         Set the number of probe points.
@@ -215,7 +224,7 @@ class Probe(Struct):
 
         return out
 
-    def __call__(self, variable):
+    def __call__(self, variable, **kwargs):
         """
         Probe the given variable. The actual implementation is in self.probe(),
         so that it can be overridden in subclasses.
@@ -224,10 +233,12 @@ class Probe(Struct):
         ----------
         variable : Variable instance
             The variable to be sampled along the probe.
+        **kwargs : additional arguments
+            See :func:`Probe.probe()`.
         """
-        return self.probe(variable)
+        return self.probe(variable, **kwargs)
 
-    def probe(self, variable):
+    def probe(self, variable, mode='val', ret_points=False):
         """
         Probe the given variable.
 
@@ -235,20 +246,29 @@ class Probe(Struct):
         ----------
         variable : Variable instance
             The variable to be sampled along the probe.
+        mode : {'val', 'grad'}, optional
+            The evaluation mode: the variable value (default) or the
+            variable value gradient.
+        ret_points : bool
+            If True, return also the probe points.
+
+        Returns
+        -------
+        pars : array
+            The parametrization of the probe points.
+        points : array, optional
+            If `ret_points` is True, the coordinates of points corresponding to
+            `pars`, where the `variable` is evaluated.
+        vals : array
+            The probed values.
         """
         refine_flag = None
 
         ev = variable.evaluate_at
-        domain = variable.field.domain
-
-        if self.share_geometry:
-            cache = domain.get_evaluate_cache(cache=Probe.cache,
-                                              share_geometry=True)
-
-        else:
-            cache = domain.get_evaluate_cache(cache=self.cache,
-                                              share_geometry=False)
+        field = variable.field
 
+        cache = field.get_evaluate_cache(cache=self.get_evaluate_cache(),
+                                         share_geometry=self.share_geometry)
         self.reset_refinement()
 
         while True:
@@ -256,11 +276,9 @@ class Probe(Struct):
             if not nm.isfinite(points).all():
                 raise ValueError('Inf/nan in probe points!')
 
-            vals, cells, status = ev(points, strategy='kdtree',
-                                     close_limit=self.options.close_limit,
-                                     cache=cache, ret_status=True)
-            ii = nm.where(status > 1)[0]
-            vals[ii] = nm.nan
+            vals, cells = ev(points, mode=mode, strategy='general',
+                             close_limit=self.options.close_limit,
+                             cache=cache, ret_cells=True)
 
             if self.is_refined:
                 break
@@ -272,7 +290,11 @@ class Probe(Struct):
 
         self.is_refined = True
 
-        return pars, vals
+        if ret_points:
+            return pars, points, vals
+
+        else:
+            return pars, vals
 
     def reset_refinement(self):
         """
@@ -345,13 +367,14 @@ class PointsProbe(Probe):
         points : array_like
             The coordinates of the points.
         """
-        points = nm.array(points, dtype=nm.float64)
+        points = nm.array(points, dtype=nm.float64, order='C')
         if points.ndim == 1:
             points.shape = points.shape + (1,)
         n_point = points.shape[0]
         name = 'points %d' % n_point
 
-        Probe.__init__(self, name=name, points=points, n_point=n_point)
+        Probe.__init__(self, name=name, share_geometry=share_geometry,
+                       points=points, n_point=n_point)
 
         self.n_point_single = n_point
 
@@ -405,7 +428,8 @@ class LineProbe(Probe):
         p1 = nm.array(p1, dtype=nm.float64)
         name = 'line [%s, %s]' % (p0, p1)
 
-        Probe.__init__(self, name=name, p0=p0, p1=p1, n_point=n_point)
+        Probe.__init__(self, name=name, share_geometry=share_geometry,
+                       p0=p0, p1=p1, n_point=n_point)
 
         dirvec = self.p1 - self.p0
         self.length = nm.linalg.norm(dirvec)
@@ -479,7 +503,8 @@ class RayProbe(Probe):
         else:
             n_point_true = n_point
 
-        Probe.__init__(self, name=name, p0=p0, dirvec=dirvec, p_fun=p_fun,
+        Probe.__init__(self, name=name, share_geometry=share_geometry,
+                       p0=p0, dirvec=dirvec, p_fun=p_fun,
                        n_point=n_point_true, both_dirs=both_dirs)
 
         self.n_point_single = n_point
@@ -551,7 +576,8 @@ class CircleProbe(Probe):
 
         name = 'circle [%s, %s, %s]' % (centre, normal, radius)
 
-        Probe.__init__(self, name=name, centre=centre, normal=normal,
+        Probe.__init__(self, name=name, share_geometry=share_geometry,
+                       centre=centre, normal=normal,
                        radius=radius, n_point=n_point)
 
     def report(self):
