@@ -1,3 +1,4 @@
+from __future__ import absolute_import
 import os
 import time
 
@@ -11,7 +12,9 @@ from sfepy.solvers.ts import TimeStepper
 from sfepy.discrete.fem.meshio import HDF5MeshIO
 from sfepy.solvers import Solver, eig
 from sfepy.linalg import MatrixAction
-from utils import iter_sym, create_pis, create_scalar_pis
+from .utils import iter_sym, iter_nonsym, create_pis, create_scalar_pis
+import six
+from six.moves import range
 
 class MiniAppBase(Struct):
     def any_from_conf(name, problem, kwargs):
@@ -26,7 +29,9 @@ class MiniAppBase(Struct):
     def __init__(self, name, problem, kwargs):
         Struct.__init__(self, name=name, problem=problem, **kwargs)
 
-        self.problem.clear_equations()
+        if self.problem is not None:
+            self.problem.clear_equations()
+
         self.set_default('requires', [])
         self.set_default('is_linear', False)
         self.set_default('dtype', nm.float64)
@@ -63,7 +68,8 @@ class MiniAppBase(Struct):
         else:
             opts = problem.conf.options
 
-        problem.set_solvers(problem.conf.solvers, opts)
+        problem.set_conf_solvers(problem.conf.solvers, opts)
+        problem.init_solvers()
 
         if self.is_linear:
             output('linear problem, trying to presolve...')
@@ -79,7 +85,7 @@ class MiniAppBase(Struct):
                 raise
 
             problem.set_linear(True)
-            problem.init_solvers(mtx=mtx_a, presolve=True)
+            problem.try_presolve(mtx_a)
 
             output('...done in %.2f s' % (time.clock() - tt))
 
@@ -112,15 +118,17 @@ class CorrMiniApp(MiniAppBase):
     def __init__(self, name, problem, kwargs):
         MiniAppBase.__init__(self, name, problem, kwargs)
         self.output_dir = self.problem.output_dir
-        self.set_default('save_name', '(not_set)')
+        self.set_default('save_name', None)
         self.set_default('dump_name', self.save_name)
         self.set_default('dump_variables', [])
         self.set_default('save_variables', self.dump_variables)
 
-        self.save_name = os.path.normpath(os.path.join(self.output_dir,
-                                                         self.save_name))
-        self.dump_name = os.path.normpath(os.path.join(self.output_dir,
-                                                         self.dump_name))
+        if self.save_name is not None:
+            self.save_name = os.path.normpath(os.path.join(self.output_dir,
+                                                           self.save_name))
+        if self.dump_name is not None:
+            self.dump_name = os.path.normpath(os.path.join(self.output_dir,
+                                                           self.dump_name))
 
     def setup_output(self, save_format=None, dump_format=None,
                       post_process_hook=None, file_per_var=None):
@@ -137,13 +145,14 @@ class CorrMiniApp(MiniAppBase):
         return self.get_save_name_base()
 
     def get_save_name(self):
-        return '.'.join((self.get_save_name_base(), self.save_format))
+        save_name_base = self.get_save_name_base()
+        if save_name_base is not None:
+            return '.'.join((self.get_save_name_base(), self.save_format))
 
     def get_dump_name(self):
-        if self.dump_format is not None:
-            return '.'.join((self.get_dump_name_base(), self.dump_format))
-        else:
-            return None
+        dump_name_base = self.get_dump_name_base()
+        if self.dump_format is not None and dump_name_base is not None:
+            return '.'.join((dump_name_base, self.dump_format))
 
     def get_output(self, corr_sol, is_dump=False, extend=True, variables=None):
         if variables is None:
@@ -170,7 +179,7 @@ class CorrMiniApp(MiniAppBase):
 
                 if is_dump:
                         var = variables[var_name]
-                        shape = (var.n_dof / var.n_components,
+                        shape = (var.n_dof // var.n_components,
                                  var.n_components)
                         out[skey] = Struct(name = 'dump', mode = 'nodes',
                                            data = dof_vector,
@@ -187,7 +196,7 @@ class CorrMiniApp(MiniAppBase):
                                                      None,
                                                      extend=extend)
 
-                    for _key, val in aux.iteritems():
+                    for _key, val in six.iteritems(aux):
                         if key:
                             new_key = _key + '_' + key
 
@@ -224,6 +233,9 @@ class ShapeDimDim(CorrMiniApp):
         corr_sol = CorrSolution(name=self.name,
                                 states=pis,
                                 components=clist)
+        self.save(corr_sol, problem,
+                  variables=problem.create_variables([self.variables[0]]))
+
         return corr_sol
 
 class ShapeDim(CorrMiniApp):
@@ -263,41 +275,39 @@ class OnesDim(CorrMiniApp):
                                 components=clist)
         return corr_sol
 
-class CopyData(CorrMiniApp):
 
+class CorrEval(CorrMiniApp):
     def __call__(self, problem=None, data=None):
         problem = get_default(problem, self.problem)
+        expr = self.expression
+        for i in range(len(self.requires)):
+            expr = expr.replace(self.requires[i],
+                                "data['%s']" % self.requires[i])
+
+        val = eval(expr)
+
         var_name = self.variable
         clist = ['data']
-        dn = self.data
 
-        if type(dn) is list:
-            data = problem
-            for ii in dn:
-                data = data.get(ii, 'None')
-        else:
-            data = problem.get(dn, 'None')
-
-        if type(data) is dict:
+        if type(val) is dict:
             corr_sol = CorrSolution(name=self.name,
-                                    state=data,
+                                    state=val,
                                     components=clist)
         else:
-            if data.dtype == nm.object:
+            if val.dtype == nm.object:
                 corr_sol = CorrSolution(name=self.name,
-                                        states=data,
+                                        states=val,
                                         components=clist)
-
-
             else:
-                ndof, ndim = data.shape
-                state = {var_name: data.reshape((ndof * ndim,))}
+                ndof, ndim = val.shape
+                state = {var_name: val.reshape((ndof * ndim,))}
                 corr_sol = CorrSolution(name=self.name,
                                         state=state,
                                         components=clist)
 
         return corr_sol
 
+
 class CorrNN(CorrMiniApp):
     """ __init__() kwargs:
         {
@@ -343,7 +353,7 @@ class CorrNN(CorrMiniApp):
                 else:
                     self.set_variables(variables, ir, ic, **data)
 
-                state = problem.solve()
+                state = problem.solve(update_materials=False)
                 assert_(state.has_ebc())
                 states[ir,ic] = state.get_parts()
 
@@ -508,7 +518,7 @@ class CorrEqPar(CorrOne):
 
         eqns ={}
         for ir in range(self.dim):
-            for key_eq, val_eq in self.equations.iteritems():
+            for key_eq, val_eq in six.iteritems(self.equations):
                 eqns[key_eq] = val_eq % self.eq_pars[ir]
 
             problem.set_equations(eqns)
@@ -559,7 +569,7 @@ class PressureEigenvalueProblem(CorrMiniApp):
         if self.mode == 'explicit':
             tt = time.clock()
             mtx_aibt = nm.zeros(mtx_bt.shape, dtype=mtx_bt.dtype)
-            for ic in xrange(mtx_bt.shape[1]):
+            for ic in range(mtx_bt.shape[1]):
                 mtx_aibt[:,ic] = ls(mtx_bt[:,ic].toarray().squeeze())
             output('mtx_aibt: %.2f s' % (time.clock() - tt))
             action_aibt = MatrixAction.from_array(mtx_aibt)
@@ -620,7 +630,7 @@ class PressureEigenvalueProblem(CorrMiniApp):
                 ii0 = get_slice(n_eigs[0], nn)
                 ii1 = get_slice(-n_eigs[1], nn)
                 eigs = nm.concatenate((eigs[ii0], eigs[ii1]))
-                mtx_q = nm.concatenate((mtx_q[:,ii0], mtx_q[:,ii1]), 1) 
+                mtx_q = nm.concatenate((mtx_q[:,ii0], mtx_q[:,ii1]), 1)
         else:
             output('required number of eigenvalues: %d' % n_eigs)
             if (n_eigs != 0) and (abs(n_eigs) < nn):
@@ -628,13 +638,6 @@ class PressureEigenvalueProblem(CorrMiniApp):
                 eigs = eigs[ii]
                 mtx_q = mtx_q[:,ii]
 
-##         from sfepy.base.plotutils import pylab, iplot
-##         pylab.semilogy(eigs)
-##         pylab.figure(2)
-##         iplot(eigs)
-##         pylab.show()
-##         debug()
-
         out = Struct(eigs=eigs, mtx_q=mtx_q)
         return out
 
@@ -850,20 +853,28 @@ class VolumeFractions(MiniAppBase):
 
         return vf
 
-class CoefSymSym(MiniAppBase):
 
+class CoefNN(MiniAppBase):
+    @staticmethod
     def set_variables_default(variables, ir, ic, mode, set_var, data):
         def get_corr_state(corr, ir, ic):
             if hasattr(corr, 'states'):
-                return corr.states[ir,ic]
-
+                if ir is None:
+                    return corr.states[ic]
+                elif ic is None:
+                    return corr.states[ir]
+                else:
+                    return corr.states[ir, ic]
             else:
                 return corr.state
 
-        mode2var = {'row' : 0, 'col' : 1}
-        idx = mode2var[mode]
+        if mode == 'row_only':
+            act_set_var = set_var
+        else:
+            mode2var = {'row': 0, 'col': 1}
+            act_set_var = [set_var[mode2var[mode]]] + set_var[2:]
 
-        for (var, req, comp) in [set_var[idx]] + set_var[2:]:
+        for (var, req, comp) in act_set_var:
             if type(req) is tuple:
                 val = get_corr_state(data[req[0]], ir, ic)[comp].copy()
                 for ii in req[1:]:
@@ -874,41 +885,67 @@ class CoefSymSym(MiniAppBase):
 
             variables[var].set_data(val)
 
-    set_variables_default = staticmethod(set_variables_default)
+    def __init__(self, name, problem, kwargs):
+        """When dim is not in kwargs, problem dimension is used."""
+        MiniAppBase.__init__(self, name, problem, kwargs)
+        self.set_default('dim', problem.get_dim())
 
-    def __call__(self, volume, problem=None, data=None):
+    def get_coef(self, row, col, volume, problem, data):
         problem = get_default(problem, self.problem)
-
-        dim, sym = problem.get_dim(get_sym=True)
-        coef = nm.zeros((sym, sym), dtype=self.dtype)
-
         term_mode = self.term_mode
         equations, variables = problem.create_evaluable(self.expression,
                                                         term_mode=term_mode)
 
-        for ir, (irr, icr) in enumerate(iter_sym(dim)):
+        coef = nm.zeros((len(row), len(col)), dtype=self.dtype)
+
+        for ir, (irr, icr) in enumerate(row):
             if isinstance(self.set_variables, list):
                 self.set_variables_default(variables, irr, icr, 'row',
                                            self.set_variables, data)
             else:
                 self.set_variables(variables, irr, icr, 'row', **data)
 
-            for ic, (irc, icc) in enumerate(iter_sym(dim)):
+            for ic, (irc, icc) in enumerate(col):
                 if isinstance(self.set_variables, list):
                     self.set_variables_default(variables, irc, icc, 'col',
                                                self.set_variables, data)
                 else:
                     self.set_variables(variables, irc, icc, 'col', **data)
 
-                val = eval_equations(equations, variables,
-                                     term_mode=term_mode)
-
-                coef[ir,ic] = val
+                val = eval_equations(equations, variables, term_mode=term_mode)
+                coef[ir, ic] = val
 
         coef /= self._get_volume(volume)
 
         return coef
 
+    def __call__(self, volume, problem=None, data=None):
+        row = [(ii, None) for ii in range(self.dim)]
+        col = [(None, ii) for ii in range(self.dim)]
+
+        return self.get_coef(row, col, volume, problem, data)
+
+
+class CoefDimDim(CoefNN):
+    pass
+
+
+class CoefSymSym(CoefNN):
+    iter_sym = staticmethod(iter_sym)
+    is_sym = True
+
+    def __call__(self, volume, problem=None, data=None):
+        problem = get_default(problem, self.problem)
+        isym = [ii for ii in self.iter_sym(problem.get_dim())]
+
+        return self.get_coef(isym, isym, volume, problem, data)
+
+
+class CoefNonSymNonSym(CoefSymSym):
+    iter_sym = staticmethod(iter_nonsym)
+    is_sym = False
+
+
 class CoefFMSymSym(MiniAppBase):
     """
     Fading memory sym x sym coefficients.
@@ -951,213 +988,71 @@ class CoefFMSymSym(MiniAppBase):
 
         return coef
 
-class CoefDimSym(MiniAppBase):
 
+class CoefDimSym(CoefNN):
     def __call__(self, volume, problem=None, data=None):
         problem = get_default(problem, self.problem)
+        dim = problem.get_dim()
+        row = [(ii, None) for ii in range(dim)]
+        col = [ii for ii in iter_sym(dim)]
 
-        dim, sym = problem.get_dim(get_sym=True)
-        coef = nm.zeros((dim, sym), dtype=self.dtype)
+        return self.get_coef(row, col, volume, problem, data)
 
-        term_mode = self.term_mode
-        equations, variables = problem.create_evaluable(self.expression,
-                                                        term_mode=term_mode)
-
-        for ir in range(dim):
-            self.set_variables(variables, ir, None, 'row', **data)
-
-            for ic, (irc, icc) in enumerate(iter_sym(dim)):
-                self.set_variables(variables, irc, icc, 'col', **data)
-
-                val = eval_equations(equations, variables,
-                                     term_mode=term_mode)
-
-                coef[ir,ic] = val
-
-        coef /= self._get_volume(volume)
-
-        return coef
-
-class CoefNN(MiniAppBase):
 
+class CoefN(CoefNN):
+    @staticmethod
     def set_variables_default(variables, ir, ic, mode, set_var, data):
-        def get_corr_state(corr, ii):
-            if hasattr(corr, 'states'):
-                return corr.states[ii]
+        mode = mode + '_only'
+        CoefNN.set_variables_default(variables, ir, ic, mode, set_var, data)
 
-            else:
-                return corr.state
-
-        mode2var = {'row' : 0, 'col' : 1}
-
-        if mode == 'col':
-            ir = ic
-        idx = mode2var[mode]
-
-        for (var, req, comp) in [set_var[idx]] + set_var[2:]:
-            if type(req) is tuple:
-                val = get_corr_state(data[req[0]], ir)[comp].copy()
-                for ii in req[1:]:
-                    val += get_corr_state(data[ii], ir)[comp]
-
-            else:
-                val = get_corr_state(data[req], ir)[comp]
-
-            variables[var].set_data(val)
-
-    set_variables_default = staticmethod(set_variables_default)
-
-    def __init__(self, name, problem, kwargs):
-        """When dim is not in kwargs, problem dimension is used."""
-        MiniAppBase.__init__(self, name, problem, kwargs)
-        self.set_default('dim', problem.get_dim())
-
-    def __call__(self, volume, problem=None, data=None):
+    def get_coef(self, row, volume, problem, data):
         problem = get_default(problem, self.problem)
-
-        coef = nm.zeros((self.dim, self.dim), dtype=self.dtype)
-
         term_mode = self.term_mode
         equations, variables = problem.create_evaluable(self.expression,
                                                         term_mode=term_mode)
 
-        if isinstance(self.set_variables, list):
-            for ir in range(self.dim):
-                self.set_variables_default(variables, ir, None, 'row',
-                                           self.set_variables, data)
-                for ic in range(self.dim):
-                    self.set_variables_default(variables, None, ic, 'col',
-                                               self.set_variables, data)
-                    val = eval_equations(equations, variables,
-                                         term_mode=term_mode)
-                    coef[ir,ic] = val
-        else:
-            for ir in range(self.dim):
-                self.set_variables(variables, ir, None, 'row', **data)
-                for ic in range(self.dim):
-                    self.set_variables(variables, None, ic, 'col', **data)
-                    val = eval_equations(equations, variables,
-                                         term_mode=term_mode)
-                    coef[ir,ic] = val
+        coef = nm.zeros((len(row),), dtype=self.dtype)
 
-        coef /= self._get_volume(volume)
-
-        return coef
-
-class CoefN(MiniAppBase):
-
-    def set_variables_default(variables, ir, set_var, data):
-        def get_corr_state(corr, ii):
-            if hasattr(corr, 'states'):
-                return corr.states[ii]
-
-            else:
-                return corr.state
-
-        for (var, req, comp) in set_var:
-            if type(req) is tuple:
-                val = get_corr_state(data[req[0]], ir)[comp].copy()
-                for ii in req[1:]:
-                    val += get_corr_state(data[ii], ir)[comp]
-
-            else:
-                val = get_corr_state(data[req], ir)[comp]
-
-            variables[var].set_data(val)
-
-    set_variables_default = staticmethod(set_variables_default)
-
-    def __init__(self, name, problem, kwargs):
-        """When dim is not in kwargs, problem dimension is used."""
-        MiniAppBase.__init__(self, name, problem, kwargs)
-        self.set_default('dim', problem.get_dim())
-
-    def __call__(self, volume, problem=None, data=None):
-        problem = get_default(problem, self.problem)
-
-        coef = nm.zeros((self.dim,), dtype=self.dtype)
-        term_mode = self.term_mode
-        equations, variables = problem.create_evaluable(self.expression,
-                                                        term_mode=term_mode)
-
-        for ir in range(self.dim):
-            if isinstance(self.set_variables, list):
-                self.set_variables_default(variables, ir, self.set_variables,
-                                           data)
-            else:
-                self.set_variables(variables, ir, **data)
-
-            val = eval_equations(equations, variables,
-                                 term_mode=term_mode)
-            coef[ir] = val
-
-        coef /= self._get_volume(volume)
-
-        return coef
-
-class CoefDimDim(CoefNN):
-    pass
-
-class CoefDim(CoefN):
-    pass
-
-class CoefSym(MiniAppBase):
-    def set_variables_default(variables, ir, ic, mode, set_var, data):
-        def get_corr_state(corr, ir, ic):
-            if hasattr(corr, 'states'):
-                return corr.states[ir,ic]
-
-            else:
-                return corr.state
-
-        if mode == 'row':
-
-            for (var, req, comp) in set_var:
-                if type(req) is tuple:
-                    val = get_corr_state(data[req[0]], ir, ic)[comp].copy()
-                    for ii in req[1:]:
-                        val += get_corr_state(data[ii], ir, ic)[comp]
-
-                else:
-                    val = get_corr_state(data[req], ir, ic)[comp]
-
-                variables[var].set_data(val)
-
-    set_variables_default = staticmethod(set_variables_default)
-
-    def __call__(self, volume, problem=None, data=None):
-        problem = get_default(problem, self.problem)
-
-        dim, sym = problem.get_dim(get_sym=True)
-        coef = nm.zeros((sym,), dtype=self.dtype)
-
-        term_mode = self.term_mode
-        equations, variables = problem.create_evaluable(self.expression,
-                                                        term_mode=term_mode)
-
-        if isinstance(self.set_variables, list):
-            self.set_variables_default(variables, None, None, 'col',
-                                       self.set_variables, data)
-
-        else:
-            self.set_variables(variables, None, None, 'col', **data)
-
-        for ii, (ir, ic) in enumerate(iter_sym(dim)):
+        for ii, (ir, ic) in enumerate(row):
             if isinstance(self.set_variables, list):
                 self.set_variables_default(variables, ir, ic, 'row',
                                            self.set_variables, data)
-
             else:
                 self.set_variables(variables, ir, ic, 'row', **data)
 
-            val = eval_equations(equations, variables,
-                                 term_mode=term_mode)
+            val = eval_equations(equations, variables, term_mode=term_mode)
             coef[ii] = val
 
         coef /= self._get_volume(volume)
 
         return coef
 
+    def __call__(self, volume, problem=None, data=None):
+        row = [(ii, None) for ii in range(self.dim)]
+
+        return self.get_coef(row, volume, problem, data)
+
+
+class CoefDim(CoefN):
+    pass
+
+
+class CoefSym(CoefN):
+    iter_sym = staticmethod(iter_sym)
+    is_sym = True
+
+    def __call__(self, volume, problem=None, data=None):
+        problem = get_default(problem, self.problem)
+        isym = [ii for ii in self.iter_sym(problem.get_dim())]
+
+        return self.get_coef(isym, volume, problem, data)
+
+
+class CoefNonSym(CoefSym):
+    iter_sym = staticmethod(iter_nonsym)
+    is_sym = False
+
+
 class CoefFMSym(MiniAppBase):
     """
     Fading memory sym coefficients.
