@@ -1,3 +1,4 @@
+from __future__ import absolute_import
 import time
 
 import numpy as nm
@@ -10,6 +11,8 @@ from sfepy.linalg import norm_l2_along_axis
 from sfepy.discrete.evaluate import eval_equations
 from sfepy.homogenization.coefs_base import MiniAppBase, CorrMiniApp
 from sfepy.homogenization.utils import coor_to_sym
+import six
+from six.moves import range
 
 def compute_eigenmomenta(em_equation, var_name, problem, eig_vectors,
                          transform=None):
@@ -24,7 +27,7 @@ def compute_eigenmomenta(em_equation, var_name, problem, eig_vectors,
     n_c = var.n_components
     eigenmomenta = nm.empty((n_eigs, n_c), dtype=nm.float64)
 
-    for ii in xrange(n_eigs):
+    for ii in range(n_eigs):
 
         if transform is None:
             vec_phi, is_zero = eig_vectors[:,ii], False
@@ -167,7 +170,7 @@ def detect_band_gaps(mass, freq_info, opts, gap_kind='normal', mtx_b=None):
     logs = [[] for ii in range(n_col + 1)]
     gaps = []
 
-    for ii in xrange(freq_info.freq_range.shape[0] + 1):
+    for ii in range(freq_info.freq_range.shape[0] + 1):
 
         f0, f1 = fm[[ii, ii+1]]
         output('interval: ]%.8f, %.8f[...' % (f0, f1))
@@ -423,6 +426,56 @@ def describe_gaps(gaps):
 
     return kinds
 
+def get_gap_ranges(freq_range, gaps, kinds):
+    """
+    For each (potential) band gap in `gaps`, return the frequency ranges of its
+    parts according to `kinds`.
+    """
+
+    def get_ranges(ii, f0, f1, kind, kind_desc, gmin, gmax):
+        if kind == 'p':
+            ranges = [(f0, f1)]
+        elif kind == 'w':
+            ranges = [(f0, f1)]
+        elif kind == 'wp':
+            ranges = [(f0, gmin[1]), (gmin[1], f1)]
+        elif kind == 's':
+            ranges = [(f0, f1)]
+        elif kind == 'sw':
+            ranges = [(f0, gmax[1]), (gmax[1], f1)]
+        elif kind == 'swp':
+            ranges = [(f0, gmax[1]), (gmax[1], gmin[1]), (gmin[1], f1)]
+        elif kind == 'is':
+            ranges = [(gmin[1], gmax[1])]
+        elif kind == 'iw':
+            ranges = [(gmin[1], gmax[1])]
+        else:
+            msg = 'impossible band gap combination! (%f %d)' % (gmin, gmax)
+            raise ValueError(msg)
+
+        return ranges
+
+    gap_ranges = []
+    for ii in range(len(freq_range) - 1):
+        f0, f1 = freq_range[[ii, ii+1]]
+        gap = gaps[ii]
+
+        if isinstance(gap, list):
+            ranges = []
+            for ig, (gmin, gmax) in enumerate(gap):
+                kind, kind_desc = kinds[ii][ig]
+                aux = get_ranges(ii, f0, f1, kind, kind_desc, gmin, gmax)
+                ranges.append(aux)
+
+        else:
+            gmin, gmax = gap
+            kind, kind_desc = kinds[ii]
+            ranges = get_ranges(ii, f0, f1, kind, kind_desc, gmin, gmax)
+
+        gap_ranges.append(ranges)
+
+    return gap_ranges
+
 def compute_cat_sym_sym(coef, iw_dir):
     """
     Christoffel acoustic tensor (part) of elasticity tensor dimension.
@@ -543,7 +596,7 @@ class SimpleEVP(CorrMiniApp):
                            dtype=nm.float64)
 
         make_full = variables.make_full_vec
-        for ii in xrange(n_eigs):
+        for ii in range(n_eigs):
             mtx_phi[:,ii] = make_full(mtx_s_phi[:,ii])
 
         return mtx_phi, mtx_phi
@@ -555,11 +608,11 @@ class SimpleEVP(CorrMiniApp):
 
         out = {}
         state = problem.create_state()
-        for ii in xrange(n_eigs):
+        for ii in range(n_eigs):
             if (ii >= save[0]) and (ii < (n_eigs - save[1])): continue
             state.set_full(mtx_phi[:,ii], force=True)
             aux = state.create_output_dict()
-            for name, val in aux.iteritems():
+            for name, val in six.iteritems(aux):
                 out[name+'%03d' % ii] = val
 
         if self.post_process_hook is not None:
@@ -589,7 +642,7 @@ class SchurEVP(SimpleEVP):
 
         mtx_b, mtx_m = mtx['B'], mtx['M']
         mtx_dib = nm.empty(mtx_b.shape, dtype=mtx_b.dtype)
-        for ic in xrange(mtx_b.shape[1]):
+        for ic in range(mtx_b.shape[1]):
             mtx_dib[:,ic] = ls(mtx_b[:,ic].toarray().squeeze())
         mtx_a = mtx['K'] + mtx_b.T * mtx_dib
 
@@ -613,7 +666,7 @@ class SchurEVP(SimpleEVP):
         mtx_s_phi_schur = - sc.dot(mtx_dib, mtx_s_phi)
         aux = nm.empty((variables.adi.ptr[-1],), dtype=nm.float64)
         set = variables.set_state_part
-        for ii in xrange(n_eigs):
+        for ii in range(n_eigs):
             set(aux, mtx_s_phi[:,ii], primary_var, stripped=True)
             set(aux, mtx_s_phi_schur[:,ii], eliminated_var,
                 stripped=True)
@@ -640,7 +693,7 @@ class DensityVolumeInfo(MiniAppBase):
         total_volume = 0.0
         volumes = {}
         densities = {}
-        for region_name, aux in self.region_to_material.iteritems():
+        for region_name, aux in six.iteritems(self.region_to_material):
             vol = vf['volume_' + region_name]
 
             mat_name, item_name = aux
@@ -679,7 +732,7 @@ class DensityVolumeInfo(MiniAppBase):
         fd.write('average density:\n')
         fd.write(ff % dv_info.average_density)
 
-        for key, val in dv_info.volumes.iteritems():
+        for key, val in six.iteritems(dv_info.volumes):
             fd.write('%s volume:\n' % key)
             fd.write(ff % val)
             fd.write('%s density:\n' % key)
@@ -908,6 +961,8 @@ class BandGaps(MiniAppBase):
         :func:`detect_band_gaps()`.
     log_save_name : str
         If not None, the band gaps log is to be saved under the given name.
+    raw_log_save_name : str
+        If not None, the raw band gaps log is to be saved under the given name.
     """
 
     def process_options(self):
@@ -929,7 +984,8 @@ class BandGaps(MiniAppBase):
                       freq_eps=get('freq_eps', 1e-8),
                       zero_eps=get('zero_eps', 1e-8),
                       detect_fun=get('detect_fun', detect_band_gaps),
-                      log_save_name=get('log_save_name', None))
+                      log_save_name=get('log_save_name', None),
+                      raw_log_save_name=get('raw_log_save_name', None))
 
     def __call__(self, volume=None, problem=None, data=None):
         problem = get_default(problem, self.problem)
@@ -976,15 +1032,19 @@ class BandGaps(MiniAppBase):
                            freq_range_margins=freq_range_margins)
 
         logs, gaps, kinds = opts.detect_fun(mass, freq_info, opts, mtx_b=mtx_b)
+        gap_ranges = get_gap_ranges(freq_range_margins, gaps, kinds)
 
         bg = Struct(name='band_gaps', logs=logs, gaps=gaps, kinds=kinds,
+                    gap_ranges=gap_ranges,
                     valid=ema.valid, eig_range=opts.eig_range,
                     n_eigs=eigs.shape[0], n_zeroed=ema.n_zeroed,
                     freq_range_initial=freq_info.freq_range_initial,
                     freq_range=freq_info.freq_range,
                     freq_range_margins=freq_info.freq_range_margins,
                     opts=opts, to_file_txt=self.to_file_txt,
-                    log_save_name=opts.log_save_name, save_log=self.save_log)
+                    log_save_name=opts.log_save_name,
+                    raw_log_save_name=opts.raw_log_save_name,
+                    save_log=self.save_log)
 
         return bg
 
@@ -1014,27 +1074,35 @@ class BandGaps(MiniAppBase):
         freq_range = bg.freq_range_margins
         fd.write('n_zeroed: %d\n' % bg.n_zeroed)
         fd.write('n_eigs: %d\n' % bg.n_eigs)
+        n_row = len(freq_range) - 1
+        fd.write('n_ranges: %d\n' % n_row)
         fd.write('f0 f1 flag_min f_min v_min flag_max f_max v_max'
                   ' kind\ndesc\n')
 
         ff = float_format
         format = "%s %s %%d %s %s %%d %s %s %%s\n%%s\n" % (6 * (ff,))
-
-        n_row = len(freq_range) - 1
-        fd.write('%d\n' % n_row)
-        for ir in xrange(n_row):
+        for ir in range(n_row):
             f0, f1 = freq_range[[ir, ir+1]]
             gmin, gmax = bg.gaps[ir]
             fd.write(format % ((f0, f1) + tuple(gmin) + tuple(gmax)
                                 + bg.kinds[ir]))
 
-        fd.write('valid resonance\n')
-        freq_range = bg.freq_range_initial
+        fd.write('\nname kind f_from f_to index f0 f1\n')
+        format = '%%s %%s %s %s %%d %s %s\n' % (4 * (ff,))
+        for ii, f0 in enumerate(bg.freq_range_margins[:-1]):
+            f1 = bg.freq_range_margins[ii + 1]
+            kind = bg.kinds[ii][0]
+            for ir, rng in enumerate(bg.gap_ranges[ii]):
+                fd.write(format
+                         % (bg.name, kind[ir], rng[0], rng[1], ii, f0, f1))
+
         n_row = len(freq_range)
-        fd.write('%d\n' % n_row)
+        fd.write('\nn_resonance: %d\n' % n_row)
+        fd.write('valid f\n')
+        freq_range = bg.freq_range_initial
         valid_in_range = bg.valid[bg.eig_range]
         format = "%%d %s\n" % ff
-        for ir in xrange(n_row):
+        for ir in range(n_row):
             fd.write(format % (valid_in_range[ir], freq_range[ir]))
         fd.close()
 
@@ -1122,7 +1190,7 @@ class PolarizationAngles(MiniAppBase):
         pas = []
 
         iw_dir = iw_dir / nla.norm(iw_dir)
-        idims = range(iw_dir.shape[0])
+        idims = list(range(iw_dir.shape[0]))
         pi2 = 0.5 * nm.pi
         for vecs in wave_vectors:
             pa = nm.empty(vecs.shape[:-1], dtype=nm.float64)
