@@ -1,19 +1,21 @@
+from __future__ import absolute_import
 import os
 
 import numpy as nm
+import six
 
 try:
     from enthought.tvtk.api import tvtk
     from enthought.mayavi.sources.vtk_data_source import VTKDataSource
+    from enthought.mayavi.core.dataset_manager import DatasetManager
     from enthought.pyface.timer.api import Timer
 
 except:
     from tvtk.api import tvtk
     from mayavi.sources.vtk_data_source import VTKDataSource
+    from mayavi.core.dataset_manager import DatasetManager
     from pyface.timer.api import Timer
 
-from dataset_manager import DatasetManager
-
 from sfepy.base.base import Struct, basestr
 from sfepy.postprocess.utils import mlab
 from sfepy.discrete.fem import Mesh
@@ -26,6 +28,17 @@ def create_file_source(filename, watch=False, offscreen=True):
     if isinstance(filename, FileSource):
         return filename
 
+    from distutils.version import LooseVersion
+
+    try:
+        from enthought.mayavi import version
+
+    except:
+        from mayavi import version
+
+    # Work around a Mayavi 4.4.x issue.
+    can_vtk_source = LooseVersion(version.version) < LooseVersion('4.4')
+
     kwargs = {'watch' : watch, 'offscreen' : offscreen}
 
     if isinstance(filename, basestr):
@@ -38,14 +51,14 @@ def create_file_source(filename, watch=False, offscreen=True):
 
     fmt = fmt.lower()
 
-    if fmt == '.vtk':
+    if can_vtk_source and (fmt == '.vtk'):
         # VTK is supported directly by Mayavi, no need to use MeshIO.
         if is_sequence:
             return VTKSequenceFileSource(filename, **kwargs)
         else:
             return VTKFileSource(filename, **kwargs)
 
-    elif fmt in supported_formats.keys():
+    elif fmt in list(supported_formats.keys()):
         if is_sequence:
             if fmt == '.h5':
                 raise ValueError('format .h5 does not support file sequences!')
@@ -195,6 +208,7 @@ class GenericFileSource(FileSource):
     def __init__(self, *args, **kwargs):
         FileSource.__init__(self, *args, **kwargs)
 
+        self.io = None
         self.read_common(self.filename)
 
     def read_common(self, filename):
@@ -313,86 +327,81 @@ class GenericFileSource(FileSource):
 
     def add_data_to_dataset(self, dataset, data):
         """Add point and cell data to the dataset."""
-        dim = self.dim
-        sym = (dim + 1) * dim / 2
-
         dm = DatasetManager(dataset=dataset)
-        for key, val in data.iteritems():
-            vd = val.data
-##             print vd.shape
+        for key, val in six.iteritems(data):
             if val.mode == 'vertex':
-                if vd.shape[1] == 1:
-                    aux = vd.reshape((vd.shape[0],))
-
-                elif vd.shape[1] == 2:
-                    zz = nm.zeros((vd.shape[0], 1), dtype=vd.dtype)
-                    aux = nm.c_[vd, zz]
-
-                elif vd.shape[1] == 3:
-                    aux = vd
-
-                else:
-                    raise ValueError('unknown vertex data format! (%s)'\
-                                     % vd.shape)
-
+                aux = self._reshape(val.data, self.dim)
                 dm.add_array(aux, key, 'point')
 
             elif val.mode == 'cell':
-                ne, aux, nr, nc = vd.shape
-                if (nr == 1) and (nc == 1):
-                    aux = vd.reshape((ne,))
-
-                elif (nr == dim) and (nc == 1):
-                    if dim == 3:
-                        aux = vd.reshape((ne, dim))
-                    else:
-                        zz = nm.zeros((vd.shape[0], 1), dtype=vd.dtype);
-                        aux = nm.c_[vd.squeeze(), zz]
-
-                elif (((nr == sym) or (nr == (dim * dim))) and (nc == 1)) \
-                         or ((nr == dim) and (nc == dim)):
-                    vd = vd.squeeze()
-
-                    if dim == 3:
-                        if nr == sym:
-                            aux = vd[:,[0,3,4,3,1,5,4,5,2]]
-                        elif nr == (dim * dim):
-                            aux = vd[:,[0,3,4,6,1,5,7,8,2]]
-                        else:
-                            aux = vd.reshape((vd.shape[0], dim*dim))
-                    else:
-                        zz = nm.zeros((vd.shape[0], 1), dtype=vd.dtype);
-                        if nr == sym:
-                            aux = nm.c_[vd[:,[0,2]], zz, vd[:,[2,1]],
-                                        zz, zz, zz, zz]
-                        elif nr == (dim * dim):
-                            aux = nm.c_[vd[:,[0,2]], zz, vd[:,[3,1]],
-                                        zz, zz, zz, zz]
-                        else:
-                            aux = nm.c_[vd[:,0,[0,1]], zz, vd[:,1,[0,1]],
-                                        zz, zz, zz, zz]
-
+                aux = self._reshape(val.data, self.dim)
                 dm.add_array(aux, key, 'cell')
 
+    @staticmethod
+    def _reshape(data, dim):
+        sym = (dim + 1) * dim // 2
+
+        if len(data.shape) == 1:
+            num, nr, nc = data.shape[0], 1, 1
+
+        elif len(data.shape) == 2:
+            num, nr, nc = data.shape[0], data.shape[1], 1
+
+        else:
+            num, _, nr, nc = data.shape
+
+        if (nr == 1) and (nc == 1):
+            aux = data.reshape((num,))
+
+        elif (nr == dim) and (nc == 1):
+            if dim == 3:
+                aux = data.reshape((num, dim))
+            else:
+                zz = nm.zeros((num, 1), dtype=data.dtype);
+                aux = nm.c_[data.squeeze(), zz]
+
+        elif ((((nr == sym) or (nr == (dim * dim))) and (nc == 1))
+              or ((nr == dim) and (nc == dim))):
+            data = data.squeeze()
+
+            if dim == 3:
+                if nr == sym:
+                    aux = data[:,[0,3,4,3,1,5,4,5,2]]
+                elif nr == (dim * dim):
+                    aux = data[:,[0,3,4,6,1,5,7,8,2]]
+                else:
+                    aux = data.reshape((num, dim*dim))
+            else:
+                zz = nm.zeros((num, 1), dtype=data.dtype);
+                if nr == sym:
+                    aux = nm.c_[data[:,[0,2]], zz, data[:,[2,1]],
+                                zz, zz, zz, zz]
+                elif nr == (dim * dim):
+                    aux = nm.c_[data[:,[0,2]], zz, data[:,[3,1]],
+                                zz, zz, zz, zz]
+                else:
+                    aux = nm.c_[data[:,0,[0,1]], zz, data[:,1,[0,1]],
+                                zz, zz, zz, zz]
+        else:
+            raise ValueError('unknown data shape! (%s)' % list(data.shape))
+
+        return aux
+
 class GenericSequenceFileSource(GenericFileSource):
     """File source usable with any format supported by MeshIO classes, with
     exception of HDF5 (.h5), for file sequences."""
 
     def read_common(self, filename):
+        GenericFileSource.read_common(self, filename[self.step])
+
         self.steps = nm.arange(len(self.filename), dtype=nm.int32)
-
-    def create_source(self):
-        """Create a VTK source from data in a SfePy-supported file."""
-        if self.io is None:
-            self.read_common(self.filename[self.step])
-
-        dataset = self.create_dataset()
-
-        src = VTKDataSource(data=dataset)
-        return src
+        self.times = nm.zeros(len(self.filename), dtype=nm.float64)
 
     def set_filename(self, filename, vis_source):
         self.filename = filename
         self.io = None
         self.source = self.create_source()
         vis_source.data = self.source.data
+
+    def file_changed(self):
+        pass
