@@ -1,7 +1,10 @@
+from __future__ import print_function
+from __future__ import absolute_import
 import os
 import shutil, tempfile
 
 import numpy as nm
+from six.moves import range
 
 try:
     from enthought.traits.api \
@@ -12,6 +15,7 @@ try:
     from enthought.tvtk.pyface.scene_editor import SceneEditor
     from enthought.mayavi.tools.mlab_scene_model import MlabSceneModel
     from enthought.mayavi.core.ui.mayavi_scene import MayaviScene
+    from enthought.mayavi.core.dataset_manager import DatasetManager
 
 except ImportError:
     from traits.api \
@@ -22,9 +26,7 @@ except ImportError:
     from tvtk.pyface.scene_editor import SceneEditor
     from mayavi.tools.mlab_scene_model import MlabSceneModel
     from mayavi.core.ui.mayavi_scene import MayaviScene
-
-
-from dataset_manager import DatasetManager
+    from mayavi.core.dataset_manager import DatasetManager
 
 from sfepy.base.base import (insert_as_static_method, output, assert_,
                              get_arguments, get_default, Struct, basestr)
@@ -404,6 +406,8 @@ class Viewer(Struct):
 
         else:
             is_subdomains = False
+            if 'mat_id' not in filter_names:
+                file_source.setup_mat_id()
 
         self.source = source = self.file_source()
         data_ranges = get_data_ranges(source, return_only=True)
@@ -473,7 +477,7 @@ class Viewer(Struct):
             for ij, iric in enumerate(cycle((n_row, n_col))):
                 ir, ic = iric
                 if ij < len(repeat):
-                    for ik in xrange(repeat[ij]):
+                    for ik in range(repeat[ij]):
                         yield ii, ir, ic
                         ii += 1
 
@@ -1233,10 +1237,10 @@ class ViewerGUI(HasTraits):
         self.scene.camera.print_traits()
         view = mlab.view()
         roll = mlab.roll()
-        print 'view:', view
-        print 'roll:', roll
-        print 'as args: --view=%.2e,%.2e,%.2e,%.2e,%.2e,%.2e --roll=%.2e' \
-              % (view[:3] + tuple(view[3]) + (roll,))
+        print('view:', view)
+        print('roll:', roll)
+        print('as args: --view=%.2e,%.2e,%.2e,%.2e,%.2e,%.2e --roll=%.2e' \
+              % (view[:3] + tuple(view[3]) + (roll,)))
 
     def _button_make_animation_steps_fired(self):
         view = mlab.view()
