@@ -1,9 +1,11 @@
+from __future__ import absolute_import
 import time
 
 import numpy as nm
 
 from sfepy.base.base import output, get_default, try_imports, Struct
 from sfepy.solvers.solvers import SolverMeta, Solver, EigenvalueSolver
+from six.moves import range
 
 def eig(mtx_a, mtx_b=None, n_eigs=None, eigenvectors=True,
         return_time=None, method='eig.scipy', **ckwargs):
@@ -64,6 +66,10 @@ class ScipyEigenvalueSolver(EigenvalueSolver):
             for dense problems :func:`eig()` or :func:`eigh()` are used, for
             sparse problems (if `n_eigs` is given) :func:`eigs()` or
             :func:`eigsh()` are used."""),
+        ('which', "'LM' | 'SM' | 'LR' | 'SR' | 'LI' | 'SI'", 'SM', False,
+         """Which eigenvectors and eigenvalues to find,
+            see :func:`scipy.sparse.linalg.eigs()`
+            or :func:`scipy.sparse.linalg.eigsh()`."""),
         ('*', '*', None, False,
          'Additional parameters supported by the method.'),
     ]
@@ -85,12 +91,16 @@ class ScipyEigenvalueSolver(EigenvalueSolver):
 
         if n_eigs is None:
             mtx_a, mtx_b = self._to_array(mtx_a, mtx_b)
-            eig = self.sla.eig if conf.method == 'eig' else self.sla.eigh
-            out = eig(mtx_a, mtx_b, right=eigenvectors, **kwargs)
+            if conf.method == 'eig':
+                out = self.sla.eig(mtx_a, mtx_b, right=eigenvectors, **kwargs)
+
+            else:
+                out = self.sla.eigh(mtx_a, mtx_b,
+                                    eigvals_only=not eigenvectors, **kwargs)
 
         else:
             eig = self.ssla.eigs if conf.method == 'eig' else self.ssla.eigsh
-            out = eig(mtx_a, M=mtx_b, k=n_eigs, which='SM',
+            out = eig(mtx_a, M=mtx_b, k=n_eigs, which=conf.which,
                       return_eigenvectors=eigenvectors, **kwargs)
 
         if eigenvectors:
@@ -121,7 +131,11 @@ class ScipySGEigenvalueSolver(EigenvalueSolver):
     def __init__(self, conf, **kwargs):
         EigenvalueSolver.__init__(self, conf, **kwargs)
 
-        import scipy.lib.lapack as llapack
+        try:
+            import scipy.linalg.lapack as llapack
+        except ImportError:
+            import scipy.lib.lapack as llapack
+
         self.llapack = llapack
 
     @standard_call
@@ -151,6 +165,10 @@ class ScipySGEigenvalueSolver(EigenvalueSolver):
         else:
             out = fun(mtx_a, mtx_b)
 
+        # Fix output order of scipy.linalg.lapack functions.
+        if out[0].ndim == 2:
+            out = (out[1], out[0]) + out[2:]
+
         if not eigenvectors:
             if n_eigs is None:
                 out = out[0]
@@ -236,13 +254,15 @@ class PysparseEigenvalueSolver(EigenvalueSolver):
          """The shifting and sorting strategy of JDSYM: strategy=0 enables the
             default JDSYM algorithm, strategy=1 enables JDSYM to avoid
             convergence to eigenvalues smaller than `tau`."""),
+        ('*', '*', None, False,
+         'Additional parameters supported by the solver.'),
     ]
 
     @staticmethod
     def _convert_mat(mtx):
         from pysparse import spmatrix
         A = spmatrix.ll_mat(*mtx.shape)
-        for i in xrange(mtx.indptr.shape[0] - 1):
+        for i in range(mtx.indptr.shape[0] - 1):
             ii = slice(mtx.indptr[i], mtx.indptr[i+1])
             n_in_row = ii.stop - ii.start
             A.update_add_at(mtx.data[ii], [i] * n_in_row, mtx.indices[ii])
@@ -260,6 +280,8 @@ class PysparseEigenvalueSolver(EigenvalueSolver):
     @standard_call
     def __call__(self, mtx_a, mtx_b=None, n_eigs=None,
                  eigenvectors=None, status=None, conf=None):
+        kwargs = self.build_solver_kwargs(conf)
+
         jdsym = self.imp['jdsym']
         itsolvers = self.imp['itsolvers']
         precon = self.imp['precon']
@@ -287,7 +309,8 @@ class PysparseEigenvalueSolver(EigenvalueSolver):
                                                 conf.eps_a, conf.i_max,
                                                 method,
                                                 clvl=conf.verbosity,
-                                                strategy=conf.strategy)
+                                                strategy=conf.strategy,
+                                                **kwargs)
 
         output('number of converged eigenvalues:', kconv, verbose=conf.verbose)
 
