@@ -1,6 +1,7 @@
 """
 Nonlinear solvers.
 """
+from __future__ import absolute_import
 import time
 
 import numpy as nm
@@ -9,6 +10,8 @@ import numpy.linalg as nla
 from sfepy.base.base import output, get_default, debug, Struct
 from sfepy.base.log import Log, get_logging_conf
 from sfepy.solvers.solvers import SolverMeta, NonlinearSolver
+import six
+from six.moves import range
 
 def check_tangent_matrix(conf, vec_x0, fun, fun_grad):
     """
@@ -226,7 +229,8 @@ class Newton(NonlinearSolver):
         eps_r = get_default(ls_eps_r, 1.0)
         lin_red = conf.eps_a * conf.lin_red
 
-        time_stats = {}
+        time_stats_keys = ['residual', 'matrix', 'solve']
+        time_stats = {key : 0.0 for key in time_stats_keys}
 
         vec_x = vec_x0.copy()
         vec_x_last = vec_x0.copy()
@@ -238,6 +242,8 @@ class Newton(NonlinearSolver):
         err = err0 = -1.0
         err_last = -1.0
         it = 0
+        ls_status = {}
+        ls_n_iter = 0
         while 1:
             if iter_hook is not None:
                 iter_hook(self, vec_x, it, err, err0)
@@ -261,7 +267,7 @@ class Newton(NonlinearSolver):
                 else:
                     ok = True
 
-                time_stats['rezidual'] = time.clock() - tt
+                time_stats['residual'] = time.clock() - tt
                 if ok:
                     try:
                         err = nla.norm(vec_r)
@@ -286,7 +292,7 @@ class Newton(NonlinearSolver):
                         break
 
                     red = conf.ls_red_warp;
-                    output('rezidual computation failed for iter %d'
+                    output('residual computation failed for iter %d'
                            ' (new ls: %e)!' % (it, red * ls))
 
                 if ls < conf.ls_min:
@@ -341,14 +347,16 @@ class Newton(NonlinearSolver):
 
             tt = time.clock()
             vec_dx = lin_solver(vec_r, x0=vec_x,
-                                eps_a=eps_a, eps_r=eps_r, mtx=mtx_a)
+                                eps_a=eps_a, eps_r=eps_r, mtx=mtx_a,
+                                status=ls_status)
+            ls_n_iter += ls_status['n_iter']
             time_stats['solve'] = time.clock() - tt
 
             if conf.verbose:
                 output('...done')
 
-            for kv in time_stats.iteritems():
-                output('%10s: %7.2f [s]' % kv)
+            for key in time_stats_keys:
+                output('%10s: %7.2f [s]' % (key, time_stats[key]))
 
             vec_e = mtx_a * vec_dx - vec_r
             lerr = nla.norm(vec_e)
@@ -365,6 +373,7 @@ class Newton(NonlinearSolver):
             status['err0'] = err0
             status['err'] = err
             status['n_iter'] = it
+            status['ls_n_iter'] = ls_n_iter if ls_n_iter >= 0 else -1
             status['condition'] = condition
 
         if conf.log.plot is not None:
@@ -443,3 +452,162 @@ class ScipyBroyden(NonlinearSolver):
             status['time_stats'] = time.clock() - tt
 
         return vec_x
+
+class PETScNonlinearSolver(NonlinearSolver):
+    """
+    Interface to PETSc SNES (Scalable Nonlinear Equations Solvers).
+
+    The solver supports parallel use with a given MPI communicator (see `comm`
+    argument of :func:`PETScNonlinearSolver.__init__()`). Returns a (global)
+    PETSc solution vector instead of a (local) numpy array, when given a PETSc
+    initial guess vector.
+
+    For parallel use, the `fun` and `fun_grad` callbacks should be provided by
+    :class:`PETScParallelEvaluator
+    <sfepy.parallel.evaluate.PETScParallelEvaluator>`.
+    """
+    name = 'nls.petsc'
+
+    __metaclass__ = SolverMeta
+
+    _parameters = [
+        ('method', 'str', 'newtonls', False,
+         'The SNES type.'),
+        ('i_max', 'int', 10, False,
+         'The maximum number of iterations.'),
+        ('if_max', 'int', 100, False,
+         'The maximum number of function evaluations.'),
+        ('eps_a', 'float', 1e-10, False,
+         'The absolute tolerance for the residual, i.e. :math:`||f(x^i)||`.'),
+        ('eps_r', 'float', 1.0, False,
+         """The relative tolerance for the residual, i.e. :math:`||f(x^i)|| /
+            ||f(x^0)||`."""),
+        ('eps_s', 'float', 0.0, False,
+         """The convergence tolerance in terms of the norm of the change in
+            the solution between steps,
+            i.e. $||delta x|| < \epsilon_s ||x||$"""),
+    ]
+
+    def __init__(self, conf, pmtx=None, prhs=None, comm=None, **kwargs):
+        if comm is None:
+            try:
+                import petsc4py
+                petsc4py.init([])
+            except ImportError:
+                msg = 'cannot import petsc4py!'
+                raise ImportError(msg)
+
+        from petsc4py import PETSc as petsc
+
+        converged_reasons = {}
+        for key, val in six.iteritems(petsc.SNES.ConvergedReason.__dict__):
+            if isinstance(val, int):
+                converged_reasons[val] = key
+
+        ksp_converged_reasons = {}
+        for key, val in six.iteritems(petsc.KSP.ConvergedReason.__dict__):
+            if isinstance(val, int):
+                ksp_converged_reasons[val] = key
+
+        NonlinearSolver.__init__(self, conf, petsc=petsc,
+                                 pmtx=pmtx, prhs=prhs, comm=comm,
+                                 converged_reasons=converged_reasons,
+                                 ksp_converged_reasons=ksp_converged_reasons,
+                                 **kwargs)
+
+    def __call__(self, vec_x0, conf=None, fun=None, fun_grad=None,
+                 lin_solver=None, iter_hook=None, status=None,
+                 pmtx=None, prhs=None, comm=None):
+        conf = self.conf
+        fun = get_default(fun, self.fun)
+        fun_grad = get_default(fun_grad, self.fun_grad)
+        lin_solver = get_default(lin_solver, self.lin_solver)
+        iter_hook = get_default(iter_hook, self.iter_hook)
+        status = get_default(status, self.status)
+        pmtx = get_default(pmtx, self.pmtx)
+        prhs = get_default(prhs, self.prhs)
+        comm = get_default(comm, self.comm)
+
+        tt = time.clock()
+
+        if isinstance(vec_x0, self.petsc.Vec):
+            psol = vec_x0
+
+        else:
+            psol = pmtx.getVecLeft()
+            psol[...] = vec_x0
+
+        snes = self.petsc.SNES()
+        snes.create(comm)
+        snes.setType(conf.method)
+
+        ksp = lin_solver.create_ksp()
+        snes.setKSP(ksp)
+        ls_conf = lin_solver.conf
+        ksp.setTolerances(atol=ls_conf.eps_a, rtol=ls_conf.eps_r,
+                          divtol=ls_conf.eps_d, max_it=ls_conf.i_max)
+
+        snes.setFunction(fun, prhs)
+        snes.setJacobian(fun_grad, pmtx)
+
+        snes.setTolerances(atol=conf.eps_a, rtol=conf.eps_r,
+                           stol=conf.eps_s, max_it=conf.i_max)
+        snes.setMaxFunctionEvaluations(conf.if_max)
+        snes.setFromOptions()
+
+        fun(snes, psol, prhs)
+        err0 = prhs.norm()
+
+        snes.solve(prhs.duplicate(), psol)
+
+        if status is not None:
+            status['time_stats'] = time.clock() - tt
+
+        if snes.reason in self.converged_reasons:
+            reason = 'snes: %s' % self.converged_reasons[snes.reason]
+
+        else:
+            reason = 'ksp: %s' % self.ksp_converged_reasons[snes.reason]
+
+        output('%s(%s): %d iterations in the last step'
+               % (ksp.getType(), ksp.getPC().getType(),
+                  ksp.getIterationNumber()),
+               verbose=conf.verbose)
+
+        output('%s convergence: %s (%s, %d iterations, %d function evaluations)'
+               % (snes.getType(), snes.reason, reason,
+                  snes.getIterationNumber(), snes.getFunctionEvaluations()),
+               verbose=conf.verbose)
+
+        converged = snes.reason >= 0
+
+        if not converged:
+            # PETSc does not update the solution if KSP have not converged.
+            dpsol = snes.getSolutionUpdate()
+            psol -= dpsol
+
+            fun(snes, psol, prhs)
+            err = prhs.norm()
+
+        else:
+            try:
+                err = snes.getFunctionNorm()
+
+            except AttributeError:
+                fun(snes, psol, prhs)
+                err = prhs.norm()
+
+        if status is not None:
+            status['err0'] = err0
+            status['err'] = err
+            status['n_iter'] = snes.getIterationNumber()
+            status['ls_n_iter'] = snes.getLinearSolveIterations()
+            status['condition'] = 0 if converged else -1
+
+        if isinstance(vec_x0, self.petsc.Vec):
+            sol = psol
+
+        else:
+            sol = psol[...].copy()
+
+        return sol
