@@ -1,7 +1,9 @@
 """
 Base (abstract) solver classes.
 """
+from __future__ import absolute_import
 from sfepy.base.base import Struct
+import six
 
 def make_get_conf(conf, kwargs):
     def _get_conf_item(name, default=None, msg_if_none=None):
@@ -132,7 +134,7 @@ class Solver(Struct):
             options = Solver._parameters + cls._parameters
 
         else:
-            options = cls._parameters
+            options = Solver._parameters
 
         opts = Struct()
         allow_extra = False
@@ -146,13 +148,13 @@ class Solver(Struct):
 
         if allow_extra:
             all_keys = set(conf.to_dict().keys())
-            other = all_keys.difference(opts.to_dict().keys())
+            other = all_keys.difference(list(opts.to_dict().keys()))
             for name in other:
                 setattr(opts, name, get(name, None, None))
 
         return opts
 
-    def __init__(self, conf=None, **kwargs):
+    def __init__(self, conf=None, context=None, **kwargs):
         if conf is None:
             conf = Struct()
 
@@ -170,7 +172,8 @@ class Solver(Struct):
                 raise ValueError('solver kind cannot be determined!')
 
         new_conf = self.process_conf(conf, kwargs)
-        Struct.__init__(self, conf=new_conf, orig_conf=conf, **kwargs)
+        Struct.__init__(self, conf=new_conf, orig_conf=conf, context=context,
+                        **kwargs)
 
     def build_solver_kwargs(self, conf):
         """
@@ -187,7 +190,7 @@ class Solver(Struct):
         std = set([ii[0] for ii in options if ii[0] != '*'])
 
         kwargs = {}
-        for key, val in conf.to_dict().iteritems():
+        for key, val in six.iteritems(conf.to_dict()):
             if key not in std:
                 kwargs[key] = val
 
@@ -196,15 +199,20 @@ class Solver(Struct):
     def __call__(self, **kwargs):
         raise ValueError('called an abstract Solver instance!')
 
+    def set_field_split(self, field_ranges, **kwargs):
+        pass
+
 class LinearSolver(Solver):
     """
     Abstract linear solver class.
     """
-    def __init__(self, conf, mtx=None, status=None, **kwargs):
-        Solver.__init__(self, conf=conf, mtx=mtx, status=status, **kwargs)
+    def __init__(self, conf, mtx=None, status=None, context=None, **kwargs):
+        Solver.__init__(self, conf=conf, mtx=mtx, status=status,
+                        context=context, **kwargs)
+        self.mtx_digest = (0, '')
 
     def __call__(self, rhs, x0=None, conf=None, eps_a=None, eps_r=None,
-                 i_max=None, mtx=None, status=None, **kwargs):
+                 i_max=None, mtx=None, status=None, context=None, **kwargs):
         raise ValueError('called an abstract LinearSolver instance!')
 
     def get_tolerance(self):
@@ -224,10 +232,10 @@ class NonlinearSolver(Solver):
     """
 
     def __init__(self, conf, fun=None, fun_grad=None, lin_solver=None,
-                 iter_hook=None, status=None, **kwargs):
+                 iter_hook=None, status=None, context=None, **kwargs):
         Solver.__init__(self, conf=conf, fun=fun, fun_grad=fun_grad,
                         lin_solver=lin_solver, iter_hook=iter_hook,
-                        status=status, **kwargs)
+                        status=status, context=context, **kwargs)
 
     def __call__(self, state0, conf=None, fun=None, fun_grad=None,
                  lin_solver=None, iter_hook=None, status=None):
@@ -238,26 +246,25 @@ class TimeSteppingSolver(Solver):
     Abstract time stepping solver class.
     """
 
-    def __init__(self, conf, **kwargs):
-        Solver.__init__(self, conf=conf, **kwargs)
+    def __init__(self, conf, nls=None, status=None, context=None, **kwargs):
+        Solver.__init__(self, conf=conf, nls=nls, status=status,
+                        context=context, **kwargs)
 
-    def __call__(self, state0=None, save_results=True, step_hook=None,
-                 post_process_hook=None, nls_status=None):
+    def __call__(self, vec0=None, nls=None, init_fun=None,
+                 prestep_fun=None, poststep_fun=None,
+                 status=None, **kwargs):
         raise ValueError('called an abstract TimeSteppingSolver instance!')
 
-    def init_time(self, nls_status=None):
-        pass
-
 class OptimizationSolver(Solver):
     """
     Abstract optimization solver class.
     """
 
     def __init__(self, conf, obj_fun=None, obj_fun_grad=None, status=None,
-                 obj_args=None,  **kwargs ):
+                 obj_args=None, context=None, **kwargs):
         Solver.__init__(self, conf=conf, obj_fun=obj_fun,
                         obj_fun_grad=obj_fun_grad, status=status,
-                        obj_args=obj_args, **kwargs)
+                        obj_args=obj_args, context=context, **kwargs)
 
     def __call__(self, state0, conf=None, obj_fun=None, obj_fun_grad=None,
                  status=None, obj_args=None):
@@ -269,10 +276,10 @@ class EigenvalueSolver(Solver):
     """
 
     def __init__(self, conf, mtx_a=None, mtx_b=None, n_eigs=None,
-                 eigenvectors=None, status=None):
+                 eigenvectors=None, status=None, context=None, **kwargs):
         Solver.__init__(self, conf=conf, mtx_a=mtx_a, mtx_b=mtx_b,
                         n_eigs=n_eigs, eigenvectors=eigenvectors,
-                        status=status)
+                        status=status, context=context)
 
     def __call__(self, mtx_a, mtx_b=None, n_eigs=None,
                  eigenvectors=None, status=None, conf=None):
