@@ -469,11 +469,10 @@ int32 dw_he_rtm( FMField *out,
                  Mapping *vg,
                  int32 isDiff, int32 mode_ul )
 {
-  int32 ii, j, sym, nRow, nQP, nEP, ret = RET_OK, dim;
+  int32 ii, j, sym, nRow, nQP, ret = RET_OK, dim;
   FMField *aux = 0, *out_qp = 0, *btd = 0, *btdb = 0, *ktsc = 0, *iktsc = 0;
 
   nQP = vg->bfGM->nLev;
-  nEP = vg->bfGM->nCol;
   sym = stress->nRow;
   nRow = out->nRow; // dim * nEP.
   dim = vg->dim;
@@ -1275,7 +1274,7 @@ int32 dq_ul_he_tan_mod_mooney_rivlin( FMField *out, FMField *mat,
                                       FMField *detF, FMField *trB,
                                       FMField *vecBS, FMField *in2B )
 {
-  int32 ii, nQP, ir, ic, iqp, sym, dim, ret = RET_OK;
+  int32 ii, nQP, ir, ic, iqp, sym, ret = RET_OK;
   float64 F1, F2, F3, F4, F5, F6, detF23, detF43;
   float64 *pd, *pkappa, *ptrace;
   float64 *ptrB, *pdetF, *pB, *pin2B, *pB_ikjl, *pB_iljk, *pd_ikjl, *pd_iljk, *pBB;
@@ -1284,7 +1283,6 @@ int32 dq_ul_he_tan_mod_mooney_rivlin( FMField *out, FMField *mat,
 
   sym = out->nRow;
   nQP = out->nLev;
-  dim = sym2dim( sym );
   ptrace = get_trace( sym );
 
   fmf_createAlloc( &B_ikjl, 1, nQP, sym, sym );
@@ -1626,7 +1624,6 @@ int32 dw_ul_volume( FMField *out, FMField *detF,
 {
   int32 ii, iqp, nQP, nEPu, nEPp, dim, ret = RET_OK;
   FMField *aux = 0, aux2[1];
-  float64 *paux, *pgc;
 
   nQP = vgv->bfGM->nLev;
   nEPu = vgv->bfGM->nCol;
@@ -2032,7 +2029,7 @@ int32 d_tl_volume_surface( FMField *out, FMField *coors,
 int32 dq_def_grad( FMField *out, FMField *state, Mapping *vg,
                    int32 *conn, int32 nEl, int32 nEP, int32 mode )
 {
-  int32 ii, iel, id, iqp, nQP, dim, ret = RET_OK;
+  int32 ii, id, iqp, nQP, dim, ret = RET_OK;
   FMField *st = 0, *mtxF = 0;
 
   state->val = FMF_PtrFirst( state );
