@@ -4,6 +4,34 @@ from sfepy.base.base import assert_
 from sfepy.linalg import dot_sequences
 from sfepy.terms.terms import Term, terms
 
+class ZeroTerm(Term):
+    r"""
+    A do-nothing term useful for introducing additional variables into the
+    equations.
+
+    :Definition:
+
+    .. math::
+        0
+
+    :Arguments:
+        - virtual  : :math:`q` or :math:`\ul{v}`
+        - state : :math:`p` or :math:`\ul{u}`
+    """
+    name = 'dw_zero'
+    arg_types = ('virtual', 'state')
+    arg_shapes = {'virtual' : ('N', None), 'state' : 'N'}
+
+    @staticmethod
+    def function(out):
+        out.fill(0.0)
+
+        return 0
+
+    def get_fargs(self, vvar, svar,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        return ()
+
 class IntegrateVolumeTerm(Term):
     r"""
     Evaluate (weighted) variable in a volume region.
@@ -292,9 +320,9 @@ class VolumeSurfaceTerm(Term):
 
     def get_fargs(self, parameter,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        ap, sg = self.get_approximation(parameter)
+        sg, _ = self.get_mapping(parameter)
 
-        sd = ap.surface_data[self.region.name]
+        sd = parameter.field.surface_data[self.region.name]
         coor = parameter.field.get_coor()
 
         return coor, sg, sd.econn.copy()
@@ -316,26 +344,27 @@ class SurfaceMomentTerm(Term):
         \int_{\Gamma} \ul{n} (\ul{x} - \ul{x}_0)
 
     :Arguments:
+        - material  : :math:`\ul{x}_0` (special)
         - parameter : any variable
-        - shift     : :math:`\ul{x}_0`
     """
-    name = 'di_surface_moment'
-    arg_types = ('parameter', 'shift')
+    name = 'd_surface_moment'
+    arg_types = ('material', 'parameter')
+    arg_shapes = {'material' : '.: D', 'parameter' : 'N'}
     integration = 'surface'
 
     function = staticmethod(terms.di_surface_moment)
 
-    def get_fargs(self, parameter, shift,
+    def get_fargs(self, material, parameter,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        ap, sg = self.get_approximation(parameter)
+        sg, _ = self.get_mapping(parameter)
 
-        sd = ap.surface_data[self.region.name]
+        sd = parameter.field.surface_data[self.region.name]
         coor = parameter.field.get_coor() \
-               - nm.asarray(shift, dtype=nm.float64)[None,:]
+               - nm.asarray(material, dtype=nm.float64)[None,:]
 
         return coor, sg, sd.econn.copy()
 
-    def get_eval_shape(self, parameter, shift,
+    def get_eval_shape(self, material, parameter,
                        mode=None, term_mode=None, diff_var=None, **kwargs):
         n_fa, n_qp, dim, n_fn, n_c = self.get_data_shape(parameter)
 
@@ -370,10 +399,7 @@ class IntegrateMatTerm(Term):
     """
     name = 'ev_integrate_mat'
     arg_types = ('material', 'parameter')
-    arg_shapes = [{'material' : '1, 1', 'parameter' : 'N'},
-                  {'material' : 'D, D'},
-                  {'material' : 'S, S'},
-                  {'material' : 'D, S'}]
+    arg_shapes = [{'material' : 'N, N', 'parameter' : 'N'}]
 
     @staticmethod
     def function(out, mat, geo, fmode):
