@@ -19,9 +19,9 @@ class DotProductVolumeTerm(Term):
         \int_\Omega c q p \mbox{ , } \int_\Omega c \ul{v} \cdot \ul{u}
         \mbox{ , }
         \int_\Omega c p r \mbox{ , } \int_\Omega c \ul{u} \cdot \ul{w} \\
-        \int_\Omega \ul{v} \cdot \ull{M} \cdot \ul{u}
+        \int_\Omega \ul{v} \cdot (\ull{M} \ul{u})
         \mbox{ , }
-        \int_\Omega \ul{u} \cdot \ull{M} \cdot \ul{w}
+        \int_\Omega \ul{u} \cdot (\ull{M} \ul{w})
 
     :Arguments 1:
         - material : :math:`c` or :math:`\ull{M}` (optional)
@@ -320,8 +320,8 @@ class VectorDotGradScalarTerm(Term):
         \int_{\Omega} \ul{u} \cdot \nabla q \\
         \int_{\Omega} c \ul{v} \cdot \nabla p \mbox{ , }
         \int_{\Omega} c \ul{u} \cdot \nabla q \\
-        \int_{\Omega} \ul{v} \cdot \ull{M} \cdot \nabla p \mbox{ , }
-        \int_{\Omega} \ul{u} \cdot \ull{M} \cdot \nabla q
+        \int_{\Omega} \ul{v} \cdot (\ull{M} \nabla p) \mbox{ , }
+        \int_{\Omega} \ul{u} \cdot (\ull{M} \nabla q)
 
     :Arguments 1:
         - material : :math:`c` or :math:`\ull{M}` (optional)
@@ -441,7 +441,7 @@ class VectorDotScalarTerm(Term):
     def dw_dot(out, mat, val_qp, bfve, bfsc, geo, fmode):
 
         nel, nqp, dim, nc = mat.shape
-        nen = bfve.shape[2]
+        nen = bfve.shape[3]
 
         status1 = 0
         if fmode in [0, 1, 3]:
@@ -481,11 +481,11 @@ class VectorDotScalarTerm(Term):
         coef = coef.reshape(coef.shape[:2] + (dim, 1))
 
         if mode == 'weak':
-            apv, vgv = self.get_approximation(vvar)
-            aps, vgs = self.get_approximation(svar)
+            vgv, _ = self.get_mapping(vvar)
+            vgs, _ = self.get_mapping(svar)
 
-            bfve = apv.get_base('v', 0, self.integral)
-            bfsc = aps.get_base('v', 0, self.integral)
+            bfve = vgv.bf
+            bfsc = vgs.bf
 
             if self.mode == 'v_weak':
                 qp_var, geo, fmode = svar, vgv, 0
@@ -545,7 +545,7 @@ class ScalarDotGradIScalarTerm(Term):
     """
     name = 'dw_s_dot_grad_i_s'
     arg_types = ('material', 'virtual', 'state')
-    arg_shapes = {'material' : '1, 1', 'virtual' : (1, 'state'), 'state' : 1}
+    arg_shapes = {'material' : '.: 1, 1', 'virtual' : (1, 'state'), 'state' : 1}
 
     @staticmethod
     def dw_fun(out, bf, vg, grad, idx, fmode):
@@ -575,13 +575,12 @@ class ScalarDotGradIScalarTerm(Term):
                 grad = nm.array([0], ndmin=4, dtype=nm.float64)
                 fmode = 1
 
-            ap, vg = self.get_approximation(virtual)
-            aps, vgs = self.get_approximation(state)
+            vg, _ = self.get_mapping(virtual)
+            vgs, _ = self.get_mapping(state)
 
-            bf = aps.get_base('v', 0, self.integral)
-            idx = int(material[0, 0, 0, 0])
+            idx = int(material)
 
-            return bf, vg, grad, idx, fmode
+            return vgs.bf, vg, grad, idx, fmode
 
         else:
             raise ValueError('unsupported evaluation mode in %s! (%s)'
@@ -589,3 +588,72 @@ class ScalarDotGradIScalarTerm(Term):
 
     def set_arg_types(self):
         self.function = self.dw_fun
+
+class ScalarDotMGradScalarTerm(Term):
+    r"""
+    Volume dot product of a scalar gradient dotted with a material vector with
+    a scalar.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} q \ul{y} \cdot \nabla p \mbox{ , }
+        \int_{\Omega} p \ul{y} \cdot \nabla q
+
+    :Arguments 1:
+        - material : :math:`\ul{y}`
+        - virtual  : :math:`q`
+        - state    : :math:`p`
+
+    :Arguments 2:
+        - material : :math:`\ul{y}`
+        - state    : :math:`p`
+        - virtual  : :math:`q`
+    """
+    name = 'dw_s_dot_mgrad_s'
+    arg_types = (('material', 'virtual', 'state'),
+                 ('material', 'state', 'virtual'))
+    arg_shapes = [{'material' : 'D, 1',
+                   'virtual/grad_state' : (1, None),
+                   'state/grad_state' : 1,
+                   'virtual/grad_virtual' : (1, None),
+                   'state/grad_virtual' : 1}]
+    modes = ('grad_state', 'grad_virtual')
+
+    @staticmethod
+    def function(out, out_qp, geo, fmode):
+        status = geo.integrate(out, out_qp)
+        return status
+
+    def get_fargs(self, mat, var1, var2,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        vg1, _ = self.get_mapping(var1)
+        vg2, _ = self.get_mapping(var2)
+
+        if diff_var is None:
+            if self.mode == 'grad_state':
+                geo = vg1
+                bf_t = vg1.bf.transpose((0, 1, 3, 2))
+                val_qp = self.get(var2, 'grad')
+                out_qp = bf_t * dot_sequences(mat, val_qp, 'ATB')
+
+            else:
+                geo = vg2
+                val_qp = self.get(var1, 'val')
+                out_qp = dot_sequences(vg2.bfg, mat, 'ATB') * val_qp
+
+            fmode = 0
+
+        else:
+            if self.mode == 'grad_state':
+                geo = vg1
+                bf_t = vg1.bf.transpose((0, 1, 3, 2))
+                out_qp = bf_t * dot_sequences(mat, vg2.bfg, 'ATB')
+
+            else:
+                geo = vg2
+                out_qp = dot_sequences(vg2.bfg, mat, 'ATB') * vg1.bf
+
+            fmode = 1
+
+        return out_qp, geo, fmode
