@@ -1,6 +1,5 @@
 import numpy as nm
 
-from sfepy.base.base import use_method_with_name, assert_
 from sfepy.linalg import dot_sequences
 from sfepy.homogenization.utils import iter_sym
 from sfepy.terms.terms import Term, terms
@@ -52,11 +51,6 @@ class LinearElasticTerm(Term):
 ##     symbolic = {'expression': expr,
 ##                 'map' : {'u' : 'state', 'D_sym' : 'material'}}
 
-    def check_shapes(self, mat, virtual, state):
-        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(state)
-        sym = (dim + 1) * dim / 2
-        assert_(mat.shape == (n_el, n_qp, sym, sym))
-
     def get_fargs(self, mat, virtual, state,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
         vg, _ = self.get_mapping(state)
@@ -95,6 +89,49 @@ class LinearElasticTerm(Term):
         else:
             self.function = terms.d_lin_elastic
 
+class LinearElasticIsotropicTerm(LinearElasticTerm):
+    r"""
+    Isotropic linear elasticity term.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u}) \mbox{ with }
+        D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
+        \lambda \ \delta_{ij} \delta_{kl}
+
+    :Arguments:
+        - material_1 : :math:`\lambda`
+        - material_2 : :math:`\mu`
+        - virtual    : :math:`\ul{v}`
+        - state      : :math:`\ul{u}`
+
+    :Arguments 2:
+        - material    : :math:`D_{ijkl}`
+        - parameter_1 : :math:`\ul{w}`
+        - parameter_2 : :math:`\ul{u}`
+    """
+    name = 'dw_lin_elastic_iso'
+    arg_types = (('material_1', 'material_2', 'virtual', 'state'),
+                 ('material_1', 'material_2', 'parameter_1', 'parameter_2'))
+    arg_shapes = {'material_1' : '1, 1', 'material_2' : '1, 1',
+                  'virtual' : ('D', 'state'), 'state' : 'D',
+                  'parameter_1' : 'D', 'parameter_2' : 'D'}
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+
+    def get_fargs(self, lam, mu, virtual, state,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        from sfepy.mechanics.matcoefs import stiffness_from_lame
+
+        mat = stiffness_from_lame(self.region.dim, lam, mu)
+        return LinearElasticTerm.get_fargs(self, mat, virtual, state,
+                                           mode=mode, term_mode=term_mode,
+                                           diff_var=diff_var, **kwargs)
+
+    def get_eval_shape(self, mat1, mat2, virtual, state,
+                       mode=None, term_mode=None, diff_var=None, **kwargs):
+        return LinearElasticTerm.get_eval_shape(self, None, None, state)
+
 class SDLinearElasticTerm(Term):
     r"""
     Sensitivity analysis of the linear elastic term.
@@ -121,55 +158,18 @@ class SDLinearElasticTerm(Term):
     arg_shapes = {'material' : 'S, S',
                   'parameter_w' : 'D', 'parameter_u' : 'D',
                   'parameter_mesh_velocity' : 'D'}
-    function = terms.d_lin_elastic
-
-    @staticmethod
-    def op_dv(vgrad):
-        nel, nlev, dim, _ = vgrad.shape
-        sd = nm.zeros((nel, nlev, dim**2, dim**2), dtype=vgrad.dtype)
-
-        if dim == 1:
-            sd[...] = vgrad[:,:]
-
-        elif dim == 2:
-            sd[:,:,0:2,0:2] = vgrad[:,:]
-            sd[:,:,2:4,2:4] = vgrad[:,:]
-
-        elif dim == 3:
-            sd[:,:,0:3,0:3] = vgrad[:,:]
-            sd[:,:,3:6,3:6] = vgrad[:,:]
-            sd[:,:,6:9,6:9] = vgrad[:,:]
-        else:
-            exit('not yet implemented!')
-
-        return sd
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+    function = terms.d_sd_lin_elastic
 
     def get_fargs(self, mat, par_w, par_u, par_mv,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
         vg, _ = self.get_mapping(par_u)
 
-        grad_w = self.get(par_w, 'grad').transpose((0,1,3,2))
-        grad_u = self.get(par_u, 'grad').transpose((0,1,3,2))
-        nel, nqp, nr, nc = grad_u.shape
-        strain_w = grad_w.reshape((nel, nqp, nr * nc, 1))
-        strain_u = grad_u.reshape((nel, nqp, nr * nc, 1))
-
-        mat_map = {1: nm.array([0]),
-                   3: nm.array([0, 2, 2, 1]),
-                   6: nm.array([0, 3, 4, 3, 1, 5, 4, 5, 2])}
-
-        mmap = mat_map[mat.shape[-1]]
-        mat_ns = mat[nm.ix_(nm.arange(nel), nm.arange(nqp),
-                            mmap, mmap)]
-
-        div_mv = self.get(par_mv, 'div')
+        grad_w = self.get(par_w, 'grad')
+        grad_u = self.get(par_u, 'grad')
         grad_mv = self.get(par_mv, 'grad')
-        opd_mv = self.op_dv(grad_mv)
 
-        aux = dot_sequences(mat_ns, opd_mv)
-        mat_mv = mat_ns * div_mv - (aux + aux.transpose((0,1,3,2)))
-
-        return 1.0, strain_w, strain_u, mat_mv, vg
+        return 1.0, grad_w, grad_u, grad_mv, mat, vg
 
     def get_eval_shape(self, mat, par_w, par_u, par_mv,
                        mode=None, term_mode=None, diff_var=None, **kwargs):
@@ -177,49 +177,6 @@ class SDLinearElasticTerm(Term):
 
         return (n_el, 1, 1, 1), par_u.dtype
 
-class LinearElasticIsotropicTerm(Term):
-    r"""
-    Isotropic linear elasticity term.
-
-    :Definition:
-
-    .. math::
-        \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u}) \mbox{ with }
-        D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
-        \lambda \ \delta_{ij} \delta_{kl}
-
-    :Arguments:
-        - material_1 : :math:`\lambda`
-        - material_2 : :math:`\mu`
-        - virtual    : :math:`\ul{v}`
-        - state      : :math:`\ul{u}`
-    """
-    name = 'dw_lin_elastic_iso'
-    arg_types = ('material_1', 'material_2', 'virtual', 'state')
-    arg_shapes = {'material_1' : '1, 1', 'material_2' : '1, 1',
-                  'virtual' : ('D', 'state'), 'state' : 'D'}
-
-    function = staticmethod(terms.dw_lin_elastic_iso)
-
-    def get_fargs(self, lam, mu, virtual, state,
-                  mode=None, term_mode=None, diff_var=None, **kwargs):
-        vg, _ = self.get_mapping(state)
-
-        if mode == 'weak':
-            if diff_var is None:
-                strain = self.get(state, 'cauchy_strain')
-                fmode = 0
-
-            else:
-                strain = nm.array([0], ndmin=4, dtype=nm.float64)
-                fmode = 1
-
-            return strain, lam, mu, vg, fmode
-
-        else:
-            raise ValueError('unsupported evaluation mode in %s! (%s)'
-                             % (self.name, mode))
-
 class LinearElasticTHTerm(THTerm):
     r"""
     Fading memory linear elastic (viscous) term. Can use derivatives.
@@ -475,7 +432,7 @@ class CauchyStrainTerm(Term):
         if mode != 'qp':
             n_qp = 1
 
-        return (n_el, n_qp, dim * (dim + 1) / 2, 1), parameter.dtype
+        return (n_el, n_qp, dim * (dim + 1) // 2, 1), parameter.dtype
 
 class CauchyStrainSTerm(CauchyStrainTerm):
     r"""
@@ -565,7 +522,7 @@ class CauchyStressTerm(Term):
         if mode != 'qp':
             n_qp = 1
 
-        return (n_el, n_qp, dim * (dim + 1) / 2, 1), parameter.dtype
+        return (n_el, n_qp, dim * (dim + 1) // 2, 1), parameter.dtype
 
 class CauchyStressTHTerm(CauchyStressTerm, THTerm):
     r"""
@@ -678,3 +635,269 @@ class CauchyStressETHTerm(CauchyStressTerm, ETHTerm):
         out = CauchyStressTerm.get_eval_shape(self, mat0, parameter, mode,
                                               term_mode, diff_var, **kwargs)
         return out
+
+class NonsymElasticTerm(Term):
+    r"""
+    Elasticity term with non-symmetric gradient. The indices of matrix
+    :math:`D_{ijkl}` are ordered as
+    :math:`[11, 12, 13, 21, 22, 23, 31, 32, 33]` in 3D and as
+    :math:`[11, 12, 21, 22]` in 2D.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} \ull{D} \nabla\ul{u} : \nabla\ul{v}
+
+    :Arguments 1:
+        - material : :math:`\ull{D}`
+        - virtual  : :math:`\ul{v}`
+        - state    : :math:`\ul{u}`
+
+    :Arguments 2:
+        - material    : :math:`\ull{D}`
+        - parameter_1 : :math:`\ul{w}`
+        - parameter_2 : :math:`\ul{u}`
+    """
+    name = 'dw_nonsym_elastic'
+    arg_types = (('material', 'virtual', 'state'),
+                 ('material', 'parameter_1', 'parameter_2'))
+    arg_shapes = {'material' : 'D2, D2', 'virtual' : ('D', 'state'),
+                  'state' : 'D', 'parameter_1' : 'D', 'parameter_2' : 'D'}
+    modes = ('weak', 'eval')
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+
+    def get_fargs(self, mat, virtual, state,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        vg, _ = self.get_mapping(state)
+
+        if mode == 'weak':
+            if diff_var is None:
+                grad = self.get(state, 'grad').transpose((0,1,3,2))
+                nel, nqp, nr, nc = grad.shape
+                grad = grad.reshape((nel,nqp,nr*nc,1))
+                fmode = 0
+
+            else:
+                grad = nm.array([0], ndmin=4, dtype=nm.float64)
+                fmode = 1
+
+            return grad, mat, vg, fmode
+
+        elif mode == 'eval':
+            grad1 = self.get(virtual, 'grad').transpose((0,1,3,2))
+            grad2 = self.get(state, 'grad').transpose((0,1,3,2))
+            nel, nqp, nr, nc = grad1.shape
+
+            return 1.0,\
+                   grad1.reshape((nel,nqp,nr*nc,1)),\
+                   grad2.reshape((nel,nqp,nr*nc,1)),\
+                   mat, vg
+
+        else:
+            raise ValueError('unsupported evaluation mode in %s! (%s)'
+                             % (self.name, mode))
+
+    def get_eval_shape(self, mat, virtual, state,
+                       mode=None, term_mode=None, diff_var=None, **kwargs):
+        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(state)
+
+        return (n_el, 1, 1, 1), state.dtype
+
+    def set_arg_types(self):
+        if self.mode == 'weak':
+            self.function = terms.dw_nonsym_elastic
+
+        else:
+            self.function = terms.d_lin_elastic
+
+def _build_wave_strain_op(vec, bf):
+    dim = len(vec)
+
+    if dim == 2:
+        n0, n1 = vec
+        nmat = nm.array([[n0, 0],
+                         [0, n1],
+                         [n1, n0]], dtype=nm.float64)
+
+    else:
+        n0, n1, n2 = vec
+        nmat = nm.array([[n0, 0, 0],
+                         [0, n1, 0],
+                         [0, 0, n2],
+                         [n1, n0, 0],
+                         [n2, 0, n0],
+                         [0, n2, n1]], dtype=nm.float64)
+
+    out = nm.einsum('ik,cqkj->cqij', nmat, bf)
+    return out
+
+def _build_cauchy_strain_op(bfg):
+    dim = bfg.shape[2]
+    if dim == 2:
+        g1, g2 = bfg[..., 0:1, :], bfg[..., 1:2, :]
+        zz = nm.zeros_like(g1)
+        out = nm.block([[g1, zz],
+                        [zz, g2],
+                        [g2, g1]])
+
+    else:
+        g1, g2, g3 = bfg[..., 0:1, :], bfg[..., 1:2, :], bfg[..., 2:3, :]
+        zz = nm.zeros_like(g1)
+        out = nm.block([[g1, zz, zz],
+                        [zz, g2, zz],
+                        [zz, zz, g3],
+                        [g2, g1, zz],
+                        [g3, zz, g1],
+                        [zz, g3, g2]])
+
+    return out
+
+class ElasticWaveTerm(Term):
+    r"""
+    Elastic dispersion term involving the wave strain :math:`g_{ij}`,
+    :math:`g_{ij}(\ul{u}) = \frac{1}{2}(u_i \kappa_j + \kappa_i u_j)`, with the
+    wave vector :math:`\ul{\kappa}`. :math:`D_{ijkl}` is given in the usual
+    matrix form exploiting symmetry: in 3D it is :math:`6\times6` with the
+    indices ordered as :math:`[11, 22, 33, 12, 13, 23]`, in 2D it is
+    :math:`3\times3` with the indices ordered as :math:`[11, 22, 12]`.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} D_{ijkl}\ g_{ij}(\ul{v}) g_{kl}(\ul{u})
+
+    :Arguments:
+        - material_1 : :math:`D_{ijkl}`
+        - material_2 : :math:`\ul{\kappa}`
+        - virtual    : :math:`\ul{v}`
+        - state      : :math:`\ul{u}`
+    """
+    name = 'dw_elastic_wave'
+    arg_types = ('material_1', 'material_2', 'virtual', 'state')
+    arg_shapes = {'material_1' : 'S, S', 'material_2' : '.: D',
+                  'virtual' : ('D', 'state'), 'state' : 'D'}
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+
+    @staticmethod
+    def function(out, out_qp, geo, fmode):
+        status = geo.integrate(out, out_qp)
+        return status
+
+    def get_fargs(self, mat, kappa, virtual, state,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        from sfepy.discrete.variables import create_adof_conn
+
+        geo, _ = self.get_mapping(state)
+
+        n_fa, n_qp, dim, n_fn, n_c = self.get_data_shape(virtual)
+
+        # Expand basis for all components.
+        bf = geo.bf
+        ebf = nm.zeros(bf.shape[:2] + (dim, n_fn * dim), dtype=nm.float64)
+        for ir in range(dim):
+            ebf[..., ir, ir*n_fn:(ir+1)*n_fn] = bf[..., 0, :]
+
+        gmat = _build_wave_strain_op(kappa, ebf)
+
+        if diff_var is None:
+            econn = state.field.get_econn('volume', self.region)
+            adc = create_adof_conn(nm.arange(state.n_dof, dtype=nm.int32),
+                                   econn, n_c, 0)
+            vals = state()[adc]
+            # Same as nm.einsum('qij,cj->cqi', gmat[0], vals)[..., None]
+            aux = dot_sequences(gmat, vals[:, None, :, None])
+            out_qp = dot_sequences(gmat, dot_sequences(mat, aux), 'ATB')
+            fmode = 0
+
+        else:
+            out_qp = dot_sequences(gmat, dot_sequences(mat, gmat), 'ATB')
+            fmode = 1
+
+        return out_qp, geo, fmode
+
+class ElasticWaveCauchyTerm(Term):
+    r"""
+    Elastic dispersion term involving the wave strain :math:`g_{ij}`,
+    :math:`g_{ij}(\ul{u}) = \frac{1}{2}(u_i \kappa_j + \kappa_i u_j)`, with the
+    wave vector :math:`\ul{\kappa}` and the elastic strain :math:`e_{ij}`.
+    :math:`D_{ijkl}` is given in the usual matrix form exploiting symmetry: in
+    3D it is :math:`6\times6` with the indices ordered as :math:`[11, 22, 33,
+    12, 13, 23]`, in 2D it is :math:`3\times3` with the indices ordered as
+    :math:`[11, 22, 12]`.
+
+    :Definition:
+
+    .. math::
+        \int_{\Omega} D_{ijkl}\ g_{ij}(\ul{v}) e_{kl}(\ul{u}) \;,
+        \int_{\Omega} D_{ijkl}\ g_{ij}(\ul{u}) e_{kl}(\ul{v})
+
+    :Arguments 1:
+        - material_1 : :math:`D_{ijkl}`
+        - material_2 : :math:`\ul{\kappa}`
+        - virtual    : :math:`\ul{v}`
+        - state      : :math:`\ul{u}`
+
+    :Arguments 2:
+        - material_1 : :math:`D_{ijkl}`
+        - material_2 : :math:`\ul{\kappa}`
+        - state      : :math:`\ul{u}`
+        - virtual    : :math:`\ul{v}`
+    """
+    name = 'dw_elastic_wave_cauchy'
+    arg_types = (('material_1', 'material_2', 'virtual', 'state'),
+                 ('material_1', 'material_2', 'state', 'virtual'))
+    arg_shapes = {'material_1' : 'S, S', 'material_2' : '.: D',
+                  'virtual' : ('D', 'state'), 'state' : 'D'}
+    geometries = ['2_3', '2_4', '3_4', '3_8']
+    modes = ('ge', 'eg')
+
+    @staticmethod
+    def function(out, out_qp, geo, fmode):
+        status = geo.integrate(out, out_qp)
+        return status
+
+    def get_fargs(self, mat, kappa, gvar, evar,
+                  mode=None, term_mode=None, diff_var=None, **kwargs):
+        from sfepy.discrete.variables import create_adof_conn
+
+        geo, _ = self.get_mapping(evar)
+
+        n_fa, n_qp, dim, n_fn, n_c = self.get_data_shape(gvar)
+
+        # Expand basis for all components.
+        bf = geo.bf
+        ebf = nm.zeros(bf.shape[:2] + (dim, n_fn * dim), dtype=nm.float64)
+        for ir in range(dim):
+            ebf[..., ir, ir*n_fn:(ir+1)*n_fn] = bf[..., 0, :]
+
+        gmat = _build_wave_strain_op(kappa, ebf)
+        emat = _build_cauchy_strain_op(geo.bfg)
+
+        if diff_var is None:
+            avar = evar if self.mode == 'ge' else gvar
+            econn = avar.field.get_econn('volume', self.region)
+            adc = create_adof_conn(nm.arange(avar.n_dof, dtype=nm.int32),
+                                   econn, n_c, 0)
+            vals = avar()[adc]
+
+            if self.mode == 'ge':
+                # Same as aux = self.get(avar, 'cauchy_strain'),
+                aux = dot_sequences(emat, vals[:, None, :, None])
+                out_qp = dot_sequences(gmat, dot_sequences(mat, aux), 'ATB')
+
+            else:
+                aux = dot_sequences(gmat, vals[:, None, :, None])
+                out_qp = dot_sequences(emat, dot_sequences(mat, aux), 'ATB')
+
+            fmode = 0
+
+        else:
+            if self.mode == 'ge':
+                out_qp = dot_sequences(gmat, dot_sequences(mat, emat), 'ATB')
+
+            else:
+                out_qp = dot_sequences(emat, dot_sequences(mat, gmat), 'ATB')
+
+            fmode = 1
+
+        return out_qp, geo, fmode
