@@ -28,8 +28,9 @@ class DivGradTerm(Term):
     name = 'dw_div_grad'
     arg_types = (('opt_material', 'virtual', 'state'),
                  ('opt_material', 'parameter_1', 'parameter_2'))
-    arg_shapes = {'opt_material' : '1, 1', 'virtual' : ('D', 'state'),
-                  'state' : 'D', 'parameter_1' : 'D', 'parameter_2' : 'D'}
+    arg_shapes = [{'opt_material' : '1, 1', 'virtual' : ('D', 'state'),
+                   'state' : 'D', 'parameter_1' : 'D', 'parameter_2' : 'D'},
+                  {'opt_material' : None}]
     modes = ('weak', 'eval')
 
     function = staticmethod(terms.term_ns_asm_div_grad)
@@ -314,7 +315,7 @@ class GradTerm(Term):
                   mode=None, term_mode=None, diff_var=None, **kwargs):
         vg, _ = self.get_mapping(parameter)
 
-        grad = self.get(parameter, 'grad')
+        grad = self.get(parameter, 'grad', integration=self.integration)
 
         fmode = {'eval' : 0, 'el_avg' : 1, 'qp' : 2}.get(mode, 1)
 
@@ -329,6 +330,12 @@ class GradTerm(Term):
 
         return (n_el, n_qp, dim, n_c), parameter.dtype
 
+class SurfaceGradTerm(GradTerm):
+    __doc__ = GradTerm.__doc__.replace('Omega', 'Gamma')
+
+    name = 'ev_surface_grad'
+    integration = 'surface_extra'
+
 class DivTerm(Term):
     r"""
     Evaluate divergence of a vector field.
@@ -369,7 +376,7 @@ class DivTerm(Term):
                   mode=None, term_mode=None, diff_var=None, **kwargs):
         vg, _ = self.get_mapping(parameter)
 
-        div = self.get(parameter, 'div')
+        div = self.get(parameter, 'div', integration=self.integration)
 
         fmode = {'eval' : 0, 'el_avg' : 1, 'qp' : 2}.get(mode, 1)
 
@@ -384,6 +391,12 @@ class DivTerm(Term):
 
         return (n_el, n_qp, 1, 1), parameter.dtype
 
+class SurfaceDivTerm(DivTerm):
+    __doc__ = DivTerm.__doc__.replace('Omega', 'Gamma')
+
+    name = 'ev_surface_div'
+    integration = 'surface_extra'
+
 class DivOperatorTerm(Term):
     r"""
     Weighted divergence term of a test function.
@@ -504,11 +517,11 @@ class PSPGCStabilizationTerm(Term):
 
     def get_fargs(self, tau, virtual, parameter, state,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        sap, svg = self.get_approximation(virtual)
-        vap, vvg = self.get_approximation(state)
+        svg, _ = self.get_mapping(virtual)
+        vvg, _ = self.get_mapping(state)
 
         val_qp = self.get(parameter, 'val')
-        conn = vap.get_connectivity(self.region, self.integration)
+        conn = state.field.get_connectivity(self.region, self.integration)
 
         if diff_var is None:
             fmode = 0
@@ -585,10 +598,10 @@ class SUPGCStabilizationTerm(Term):
 
     def get_fargs(self, delta, virtual, parameter, state,
                   mode=None, term_mode=None, diff_var=None, **kwargs):
-        ap, vg = self.get_approximation(virtual)
+        vg, _ = self.get_mapping(virtual)
 
         val_qp = self.get(parameter, 'val')
-        conn = ap.get_connectivity(self.region, self.integration)
+        conn = virtual.field.get_connectivity(self.region, self.integration)
 
         if diff_var is None:
             fmode = 0
